package at.cl.atp.foresolution.model;

import java.util.HashMap;
import java.util.Map;

import at.cl.atp.foresolution.model.utils.BinaryKeyMap;
import at.cl.atp.foresolution.model.utils.MultikeyMap;

public class TermRegistry {

	private static final Map<Formula, NegatedFormula> negatedFormulaMapping = new HashMap<Formula, NegatedFormula>();
	private static final BinaryKeyMap<String, Formula, QuantifiedFormula> existentialQuantifierMapping = new BinaryKeyMap<String, Formula, QuantifiedFormula>();
	private static final BinaryKeyMap<String, Formula, QuantifiedFormula> universalQuantifierMapping = new BinaryKeyMap<String, Formula, QuantifiedFormula>();
	private static final Map<Integer, Map<String, MultikeyMap<Formula, Functor>>> predicateMappings = new HashMap<Integer, Map<String, MultikeyMap<Formula, Functor>>>();
	private static final Map<Integer, Map<String, MultikeyMap<Formula, Functor>>> functionMappings = new HashMap<Integer, Map<String, MultikeyMap<Formula, Functor>>>();
	private static final Map<String, Functor> nullaryFunctionMapping = new HashMap<String, Functor>();
	private static final BinaryKeyMap<Formula, Formula, BinaryFormula> disjunctionMapping = new BinaryKeyMap<Formula, Formula, BinaryFormula>();
	private static final BinaryKeyMap<Formula, Formula, BinaryFormula> conjunctionMapping = new BinaryKeyMap<Formula, Formula, BinaryFormula>();
	private static final Map<String, Atom> constantMapping = new HashMap<String, Atom>();
	private static final Map<String, Atom> variableMapping = new HashMap<String, Atom>();
	private static final Map<Long, Formula> idMapping = new HashMap<Long, Formula>();

	private TermRegistry() {
	}

	public static Formula getByID(final long id) {
		Formula returnValue = idMapping.get(id);
		if (returnValue == null) {
			throw new RuntimeException("No formula registered with ID " + id + "!");
		}
		return returnValue;
	}

	public static Long[] toIDs(final Formula... formulas) {
		Long[] returnValue = new Long[formulas.length];
		for (int i = 0; i < formulas.length; i++) {
			returnValue[i] = formulas[i].getFormulaID();
		}
		return returnValue;
	}

	public static Formula[] toFormulas(final long... argumentIDs) {
		Formula[] returnValue = new Formula[argumentIDs.length];
		for (int i = 0; i < argumentIDs.length; i++) {
			returnValue[i] = getByID(argumentIDs[i]);
		}
		return returnValue;
	}

	public static Atom createVariable(final String name) {
		Atom variable = variableMapping.get(name);
		if (variable != null) {
			return variable;
		} else {
			Atom newAtom = FormulaFactory.createVariable(name);
			variableMapping.put(name, newAtom);
			idMapping.put(newAtom.getFormulaID(), newAtom);
			return newAtom;
		}
	}

	public static Atom createConstant(final String name) {
		Atom constant = constantMapping.get(name);
		if (constant != null) {
			return constant;
		} else {
			Atom newAtom = FormulaFactory.createConstant(name);
			constantMapping.put(name, newAtom);
			idMapping.put(newAtom.getFormulaID(), newAtom);
			return newAtom;
		}
	}

	public static BinaryFormula createConjunction(final Formula left, final Formula right) {
		BinaryFormula result = conjunctionMapping.get(left, right);
		if (result != null) {
			return result;
		} else {
			result = FormulaFactory.createConjunction(left, right);
			conjunctionMapping.put(left, right, result);
			idMapping.put(result.getFormulaID(), result);
			return result;
		}
	}

	public static BinaryFormula createDisjunction(final Formula left, final Formula right) {
		BinaryFormula result = disjunctionMapping.get(left, right);
		if (result != null) {
			return result;
		} else {
			result = FormulaFactory.createDisjunction(left, right);
			disjunctionMapping.put(left, right, result);
			idMapping.put(result.getFormulaID(), result);
			return result;
		}
	}

	public static BinaryFormula createConjunction(final long leftID, final long rightID) {
		return createConjunction(getByID(leftID), getByID(rightID));
	}

	public static BinaryFormula createDisjunction(final long leftID, final long rightID) {
		return createDisjunction(getByID(leftID), getByID(rightID));
	}

	public static Functor createFunction(final String name, final Formula... arguments) {
		Functor functor = null;
		if (arguments.length == 0) {
			functor = nullaryFunctionMapping.get(name);
			if (functor != null) {
				return functor;
			} else {
				functor = FormulaFactory.createFunction(name, arguments);
				nullaryFunctionMapping.put(name, functor);
				idMapping.put(functor.getFormulaID(), functor);
				return functor;
			}
		} else {
			Map<String, MultikeyMap<Formula, Functor>> naryFunctionMap = functionMappings.get(arguments.length);
			if (naryFunctionMap == null) {
				naryFunctionMap = new HashMap<String, MultikeyMap<Formula, Functor>>();
				functionMappings.put(arguments.length, naryFunctionMap);
			}
			MultikeyMap<Formula, Functor> functionNameMap = naryFunctionMap.get(name);
			if (functionNameMap == null) {
				functionNameMap = new MultikeyMap<Formula, Functor>(arguments.length);
				naryFunctionMap.put(name, functionNameMap);
			}
			functor = functionNameMap.get(arguments);
			if (functor == null) {
				functor = FormulaFactory.createFunction(name, arguments);
				functionNameMap.put(functor, arguments);
				idMapping.put(functor.getFormulaID(), functor);
			}
			return functor;
		}
	}

	public static Functor createPredicate(final String name, final Formula... arguments) {
		Functor functor = null;
		if (arguments.length == 0) {
			throw new RuntimeException("Cannot create predicate with 0 arguments!");
		} else {
			Map<String, MultikeyMap<Formula, Functor>> naryPredicateMap = predicateMappings.get(arguments.length);
			if (naryPredicateMap == null) {
				naryPredicateMap = new HashMap<String, MultikeyMap<Formula, Functor>>();
				predicateMappings.put(arguments.length, naryPredicateMap);
			}
			MultikeyMap<Formula, Functor> predicateNameMap = naryPredicateMap.get(name);
			if (predicateNameMap == null) {
				predicateNameMap = new MultikeyMap<Formula, Functor>(arguments.length);
				naryPredicateMap.put(name, predicateNameMap);
			}
			functor = predicateNameMap.get(arguments);
			if (functor == null) {
				functor = FormulaFactory.createPredicate(name, arguments);
				predicateNameMap.put(functor, arguments);
				idMapping.put(functor.getFormulaID(), functor);
			}
			return functor;
		}
	}

	public static Functor createFunction(final String name, final long... argumentIDs) {
		return createFunction(name, toFormulas(argumentIDs));
	}

	public static Functor createPredicate(final String name, final long... argumentIDs) {
		return createPredicate(name, toFormulas(argumentIDs));
	}

	public static QuantifiedFormula createUniversalFormula(final Formula innerFormula, final String quantifiedVariable) {
		QuantifiedFormula formula = universalQuantifierMapping.get(quantifiedVariable, innerFormula);
		if (formula != null) {
			return formula;
		} else {
			formula = FormulaFactory.createUniversalFormula(innerFormula, quantifiedVariable);
			universalQuantifierMapping.put(quantifiedVariable, innerFormula, formula);
			idMapping.put(formula.getFormulaID(), innerFormula);
			return formula;
		}
	}

	public static QuantifiedFormula createExistentialFormula(final Formula innerFormula, final String quantifiedVariable) {
		QuantifiedFormula formula = existentialQuantifierMapping.get(quantifiedVariable, innerFormula);
		if (formula != null) {
			return formula;
		} else {
			formula = FormulaFactory.createExistentialFormula(innerFormula, quantifiedVariable);
			existentialQuantifierMapping.put(quantifiedVariable, innerFormula, formula);
			idMapping.put(formula.getFormulaID(), innerFormula);
			return formula;
		}
	}

	public static QuantifiedFormula createUniversalFormula(final long innerFormulaID, final String quantifiedVariable) {
		return createUniversalFormula(getByID(innerFormulaID), quantifiedVariable);
	}

	public static QuantifiedFormula createExistentialFormula(final long innerFormulaID, final String quantifiedVariable) {
		return createExistentialFormula(getByID(innerFormulaID), quantifiedVariable);
	}

	public static QuantifiedFormula createUniversalFormula(final Formula innerFormula,
			final String... quantifiedVariables) {
		QuantifiedFormula formula = null;
		for (int i = quantifiedVariables.length - 1; i >= 0; i--) {
			String variable = quantifiedVariables[i];
			if (formula == null) {
				formula = TermRegistry.createUniversalFormula(innerFormula, variable);
			} else {
				formula = TermRegistry.createUniversalFormula(formula, variable);
			}
		}
		return formula;
	}

	public static QuantifiedFormula createExistentialFormula(final Formula innerFormula,
			final String... quantifiedVariables) {
		QuantifiedFormula formula = null;
		for (int i = quantifiedVariables.length - 1; i >= 0; i--) {
			String variable = quantifiedVariables[i];
			if (formula == null) {
				formula = TermRegistry.createUniversalFormula(innerFormula, variable);
			} else {
				formula = TermRegistry.createUniversalFormula(formula, variable);
			}
		}
		return formula;
	}

	public static NegatedFormula createNegatedFormula(final Formula innerFormula) {
		NegatedFormula formula = negatedFormulaMapping.get(innerFormula);
		if (formula != null) {
			return formula;
		} else {
			formula = FormulaFactory.createNegatedFormula(innerFormula);
			negatedFormulaMapping.put(innerFormula, formula);
			idMapping.put(formula.getFormulaID(), innerFormula);
			return formula;
		}
	}

	public static QuantifiedFormula createExistentialFormula(final Formula innerFormula,
			final Iterable<String> variables) {
		QuantifiedFormula formula = null;
		for (String variable : variables) {
			if (formula == null) {
				formula = TermRegistry.createExistentialFormula(innerFormula, variable);
			} else {
				formula = TermRegistry.createExistentialFormula(formula, variable);
			}
		}
		return formula;
	}

	public static QuantifiedFormula createUniversalFormula(final Formula innerFormula, final Iterable<String> variables) {
		QuantifiedFormula formula = null;
		for (String variable : variables) {
			if (formula == null) {
				formula = TermRegistry.createUniversalFormula(innerFormula, variable);
			} else {
				formula = TermRegistry.createUniversalFormula(formula, variable);
			}
		}
		return formula;
	}
}
