package at.cl.atp.foresolution.model.substitution;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import at.cl.atp.foresolution.exceptions.NonFreeSubstitutionException;
import at.cl.atp.foresolution.exceptions.OccursCheckException;
import at.cl.atp.foresolution.exceptions.ReplacementOverrideException;
import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.BinaryFormula;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.Functor;
import at.cl.atp.foresolution.model.NegatedFormula;
import at.cl.atp.foresolution.model.QuantifiedFormula;
import at.cl.atp.foresolution.model.TermRegistry;
import at.cl.atp.foresolution.model.enums.AtomType;
import at.cl.atp.foresolution.model.enums.BinaryOperation;
import at.cl.atp.foresolution.model.enums.FunctorType;
import at.cl.atp.foresolution.model.enums.Quantifier;
import at.cl.atp.foresolution.model.utils.FormulaHelper;

public class Substitution {

	private final Map<String, Formula> replacements;

	/**
	 * Main constructor. Creates an empty substitution.<br>
	 * Use {@link Substitution#addReplacement(String, Formula)} to add variable
	 * replacements to this substitution.
	 */
	public Substitution() {
		this.replacements = new HashMap<String, Formula>();
	}

	/**
	 * Returns the replacements map from this substitution.<br>
	 * You should treat the returned map as <i>read-only</i> and modify this
	 * substitution through {@link Substitution#addReplacement(String, Formula)}
	 * or {@link Substitution#addAllReplacementsFrom(Substitution)}, because
	 * these methods include safety checks.<br>
	 * If you do have to modify the map returned by this method, be aware that
	 * it will modify the internal state of this substitution.
	 * 
	 * @return The internal replacement map of this substitution
	 * @see Substitution#addReplacement(String, Formula)
	 * @see Substitution#addAllReplacementsFrom(Substitution)
	 */
	public Map<String, Formula> getReplacements() {
		return this.replacements;
	}

	/**
	 * Adds a replacement for a single variable to this substitution.<br>
	 * More formally, this adds a mapping of the form { x -> f }, where x is the
	 * variable and f is the formula by which x should be replaced when this
	 * substitution is applied.
	 * 
	 * @param variable
	 *            The variable to replace
	 * @param target
	 *            The formula to replace the variable with
	 * @throws OccursCheckException
	 *             Thrown if the given variable occurs inside the target
	 *             formula.
	 * @throws ReplacementOverrideException
	 *             Thrown if there is already a different replacement for the
	 *             given variable.
	 */
	public void addReplacement(final String variable, final Formula target) throws OccursCheckException,
			ReplacementOverrideException {
		if (FormulaHelper.occursCheck(variable, target)) {
			throw new OccursCheckException(variable, target);
		}
		if (this.overridesCheck(variable, target)) {
			throw new ReplacementOverrideException(variable, this.replacements.get(variable), target);
		}
		this.replacements.put(variable, target);
	}

	/**
	 * Checks if adding another replacement for the given variable would
	 * override an already existing replacement.
	 * 
	 * @param variable
	 *            The new replacement variable
	 * @param formula
	 *            The formula which should replace the variable
	 * @return <code>true</code> if there already is a replacement for the
	 *         variable which is not syntactically equal to the given formula,
	 *         otherwise <code>false</code>.
	 */
	private boolean overridesCheck(final String variable, final Formula formula) {
		if (this.replacements.containsKey(variable)) {
			Formula replacement = this.replacements.get(variable);
			return replacement.contentEquals(formula) == false;
		} else {
			return false;
		}
	}

	/**
	 * Applies the substitution to the given formula.<br>
	 * This includes a check if the formula is free for the substitution.
	 * 
	 * @param formula
	 *            The formula that should be modified by applying this
	 *            substitution to it
	 * @return The formula after applying this substitution to it
	 * @throws NonFreeSubstitutionException
	 *             Thrown if this substitution specifies a variable replacement
	 *             but that variable is bound in the formula.
	 */
	public Formula applyTo(final Formula formula) throws NonFreeSubstitutionException {
		if (this.isEmpty()) {
			return formula;
		}
		if (this.isFreeForThis(formula) == false) {
			throw new NonFreeSubstitutionException(formula, this);
		}
		// System.out.println("SUBS :: " + this.toString() + "   to   " +
		// FormulaFormatter.format(formula));
		return this.applyToRec(formula);
	}

	private Formula applyToRec(final Formula formula) {
		if (formula instanceof Atom) {
			Atom atom = (Atom) formula;
			if (atom.getType() == AtomType.VARIABLE) {
				Formula replacement = this.replacements.get(atom.getName());
				if (replacement != null) {
					return replacement;
				} else {
					return atom;
				}
			} else {
				return atom;
			}
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			if (f.getQuantifier() == Quantifier.UNIVERSAL) {
				return TermRegistry.createUniversalFormula(this.applyToRec(f.getInnerFormula()),
						f.getQuantifiedVariable());
			} else {
				return TermRegistry.createExistentialFormula(this.applyToRec(f.getInnerFormula()),
						f.getQuantifiedVariable());
			}
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			Formula left = this.applyToRec(f.getLeft());
			Formula right = this.applyToRec(f.getRight());
			if (f.getOperation() == BinaryOperation.AND) {
				return TermRegistry.createConjunction(left, right);
			} else {
				return TermRegistry.createDisjunction(left, right);
			}
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			return TermRegistry.createNegatedFormula(this.applyToRec(f.getInnerFormula()));
		} else if (formula instanceof Functor) {
			Functor f = (Functor) formula;
			int argumentCount = f.getArguments().length;
			Formula[] newArgs = new Formula[argumentCount];
			for (int index = 0; index < argumentCount; index++) {
				newArgs[index] = this.applyToRec(f.getArguments()[index]);
			}
			if (f.getType() == FunctorType.FUNCTION) {
				return TermRegistry.createFunction(f.getName(), newArgs);
			} else {
				return TermRegistry.createPredicate(f.getName(), newArgs);
			}
		}
		// this place is unreachable anyways
		throw new RuntimeException();
	}

	/**
	 * Checks if the given formula is free for this substitution.<br>
	 * This is done by checking if <b>none</b> of the variables which are going
	 * to be substituted by this Substitution are bound by Quantors in the
	 * formula.
	 * 
	 * @param formula
	 *            The formula to check if it is free for this substitution.
	 * @return <code>true</code> if the formula is free for this substitution,
	 *         otherwise <code>false</code>.
	 */
	public boolean isFreeForThis(final Formula formula) {
		Set<String> boundVariables = FormulaHelper.getBoundVariables(formula);
		for (String variable : this.replacements.keySet()) {
			if (boundVariables.contains(variable)) {
				return false;
			}
		}
		// all variables in the formula are free for this substitution
		return true;
	}

	/**
	 * Checks if this is the empty substitution.
	 * 
	 * @return <code>true</code> if this is the empty substitution, otherwise
	 *         <code>false</code>.
	 */
	public boolean isEmpty() {
		return this.replacements.isEmpty();
	}

	/**
	 * Constructs the union of this and the 'other' substitution.<br>
	 * Modifies this substitution by adding all of the replacements specified in
	 * the 'other' substitution. Returns the modified version of this
	 * substitution.
	 * 
	 * @param other
	 *            The substitution which should be united with this substitution
	 * @return This substitution, modified as stated above.
	 */
	public Substitution union(final Substitution other) {
		if (other == null || other.isEmpty()) {
			return this;
		}
		this.replacements.putAll(other.replacements);
		return this;
	}

	/**
	 * Adds all replacement from the other substitution to this substitution.<br>
	 * This effectively merges the other substitution into this substitution,
	 * modifying this substitution but leaving the other substitution untouched.
	 * 
	 * @param other
	 *            The other substitution to merge this one with
	 * @throws ReplacementOverrideException
	 *             Thrown if a variable from the other substitution maps to a
	 *             different formula than declared in this substitution.
	 */
	public void addAllReplacementsFrom(final Substitution other) throws ReplacementOverrideException {
		for (Entry<String, Formula> replacement : other.replacements.entrySet()) {
			this.addReplacement(replacement.getKey(), replacement.getValue());
		}
	}

	@Override
	public String toString() {
		if (this.isEmpty()) {
			return "{ }";
		}
		StringBuilder builder = new StringBuilder();
		builder.append("{");
		for (Entry<String, Formula> replacement : this.replacements.entrySet()) {
			builder.append(replacement.getKey());
			builder.append(" -> (");
			builder.append(replacement.getValue().toString());
			builder.append(")");
			builder.append(", ");
		}
		// remove the last ", " sequence
		builder.delete(builder.length() - 2, builder.length());
		builder.append(" }");
		return builder.toString();
	}

	/**
	 * Determines if this substitution conflicts with the other substitution.<br>
	 * A conflict is defined as a replacement from this substitution and a
	 * replacement from the other substitution mapping the same variable to
	 * different target formulas.<br>
	 * In other words, if this method returns <code>false</code>, then it is
	 * possible to build a common substitution which contains all replacments
	 * from this substitution plus all replacements from the other substitution.
	 * 
	 * @param other
	 *            The other substitution to check if this conflicts with it
	 * @return <code>true</code> if there is a mapping conflict, otherwise
	 *         <code>false</code>
	 */
	public boolean conflictsWith(final Substitution other) {
		for (Entry<String, Formula> myReplacement : this.replacements.entrySet()) {
			for (Entry<String, Formula> yourReplacement : other.replacements.entrySet()) {
				if (myReplacement.getKey().contentEquals(yourReplacement.getKey())) {
					// both substitutions map the same variable
					Formula myTarget = myReplacement.getValue();
					Formula yourTarget = yourReplacement.getValue();
					if (!myTarget.contentEquals(yourTarget)) {
						// variable is assigned to different targets
						return true;
					}
				}
			}
		}
		return false;
	}

}
