package at.cl.atp.foresolution.model.subsumption;

import java.util.HashSet;
import java.util.Set;

import at.cl.atp.foresolution.model.clause.ClauseLiteral;
import at.cl.atp.foresolution.model.substitution.Substitution;

public class SubsumptionSearchTreeNode {

	private SubsumptionSearchTreeNode parent;
	private final Set<ClauseLiteral> unmatchedLiterals;
	private final Set<LiteralMatch> matches;
	private LiteralMatch chosenMatch;

	public SubsumptionSearchTreeNode() {
		this.matches = new HashSet<LiteralMatch>();
		this.unmatchedLiterals = new HashSet<ClauseLiteral>();
		this.parent = null;
		this.chosenMatch = null;
	}

	public SubsumptionSearchTreeNode getParent() {
		return this.parent;
	}

	public void setParent(final SubsumptionSearchTreeNode parent) {
		this.parent = parent;
	}

	public LiteralMatch getChosenMatch() {
		return this.chosenMatch;
	}

	public void setChosenMatch(final LiteralMatch chosenMatch) {
		this.chosenMatch = chosenMatch;
	}

	public Set<ClauseLiteral> getUnmatchedLiterals() {
		return this.unmatchedLiterals;
	}

	public Set<LiteralMatch> getMatches() {
		return this.matches;
	}

	public boolean isLeaf() {
		return this.unmatchedLiterals.isEmpty();
	}

	public boolean isError() {
		return !this.unmatchedLiterals.isEmpty() && this.matches.isEmpty();
	}

	public boolean isRoot() {
		return this.parent == null;
	}

	/**
	 * Collects all substitutions from all search tree nodes on this path to the
	 * root.
	 * 
	 * @return The set of substitutions
	 */
	private Set<Substitution> gatherSubstitutionsOnPathToRoot() {
		Set<Substitution> resultSet = new HashSet<Substitution>();
		this.gatherSubstitutionsOnPathToRootRec(resultSet);
		return resultSet;
	}

	private void gatherSubstitutionsOnPathToRootRec(final Set<Substitution> resultSet) {
		if (this.isRoot()) {
			return;
		}
		resultSet.add(this.chosenMatch.getSigma());
		this.parent.gatherSubstitutionsOnPathToRootRec(resultSet);
	}

	/**
	 * Returns the set of LiteralMatches that have the given literal as left
	 * literal.
	 * 
	 * @param literal
	 *            The literal to search for
	 * @return The set of LiteralMatches that have the given literal as left
	 *         literal.
	 */
	private Set<LiteralMatch> getMatchesFor(final ClauseLiteral literal) {
		Set<LiteralMatch> resultSet = new HashSet<LiteralMatch>();
		for (LiteralMatch candidate : this.matches) {
			if (candidate.getLeft() != literal) {
				continue;
			}
			resultSet.add(candidate);
		}
		return resultSet;
	}

	/**
	 * Recursively expands all child nodes to find a substitution sigma that
	 * transforms all ClauseLiterals into unmatched literals, using the
	 * available matches.<br>
	 * Each unmatched literal may only be matched once.
	 * 
	 * @return The substitution which is the substitution for the subsumption
	 *         property, or <code>null</code> if none exists.
	 */
	public Substitution expandChildren() {
		if (this.isError()) {
			return null;
		}
		if (this.isLeaf()) {
			Set<Substitution> substitutions = this.gatherSubstitutionsOnPathToRoot();
			Substitution sigma = new Substitution();
			for (Substitution sub : substitutions) {
				sigma.addAllReplacementsFrom(sub);
			}
			return sigma;
		}
		for (ClauseLiteral literal : this.unmatchedLiterals) {
			Set<LiteralMatch> potentialMatches = this.getMatchesFor(literal);
			for (LiteralMatch match : potentialMatches) {
				SubsumptionSearchTreeNode child = new SubsumptionSearchTreeNode();
				Set<LiteralMatch> remainingMatches = new HashSet<LiteralMatch>();
				for (LiteralMatch remainingMatch : this.matches) {
					if (remainingMatch == match) {
						// we've chosen this match, we cannot use it again
						continue;
					}
					if (remainingMatch.getRight() == match.getRight()) {
						// we cannot match to the same righthand side twice
						continue;
					}
					if (remainingMatch.getSigma().conflictsWith(match.getSigma())) {
						// conflicting substitutions will not produce a
						// subsumption
						continue;
					}
					// we keep the match
					remainingMatches.add(remainingMatch);
				}
				child.setChosenMatch(match);
				child.getMatches().addAll(remainingMatches);
				child.getUnmatchedLiterals().addAll(this.unmatchedLiterals);
				child.getUnmatchedLiterals().remove(match.getLeft());
				child.setParent(this);
				Substitution sigma = child.expandChildren();
				if (sigma != null) {
					return sigma;
				}
			}
		}
		return null;
	}

}
