package at.cl.atp.foresolution.model.unification;

import at.cl.atp.foresolution.exceptions.EqualityToSubstitutionException;
import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.Functor;
import at.cl.atp.foresolution.model.enums.AtomType;
import at.cl.atp.foresolution.model.substitution.Substitution;
import at.cl.atp.foresolution.model.utils.FormulaHelper;

public class Equality {

	private Formula left;
	private Formula right;

	public Equality(final Formula left, final Formula right) {
		this.left = left;
		this.right = right;
	}

	public Formula getLeft() {
		return this.left;
	}

	public Atom getLeftAsVariable() {
		return (Atom) this.left;
	}

	public Formula getRight() {
		return this.right;
	}

	public boolean hasVariableOnEitherSide() {
		return this.hasVariableOnLeftHandSide() || this.hasVariableOnRightHandSide();
	}

	public boolean hasVariableOnLeftHandSide() {
		return this.left instanceof Atom && ((Atom) this.left).getType() == AtomType.VARIABLE;
	}

	public boolean hasVariableOnRightHandSide() {
		return this.right instanceof Atom && ((Atom) this.right).getType() == AtomType.VARIABLE;
	}

	public boolean isSyntacticEquality() {
		return this.left.contentEquals(this.right);
	}

	public boolean hasLeadingFunction() {
		if (this.left instanceof Functor && this.right instanceof Functor) {
			return true;
		}
		return false;
	}

	public boolean isLeadingFunctionSymbolEqual() {
		if (this.left instanceof Functor && this.right instanceof Functor) {
			String leftName = ((Functor) this.left).getName();
			String rightName = ((Functor) this.right).getName();
			return leftName.contentEquals(rightName);
		} else {
			return false;
		}
	}

	public Formula[] getArgumentsOfLeftFunction() {
		if (this.left instanceof Functor) {
			Functor f = (Functor) this.left;
			return f.getArguments();
		}
		return null;
	}

	public Formula[] getArgumentsOfRightFunction() {
		if (this.right instanceof Functor) {
			Functor f = (Functor) this.right;
			return f.getArguments();
		}
		return null;
	}

	public void swapLeftAndRight() {
		Formula temp = this.left;
		this.left = this.right;
		this.right = temp;
	}

	public boolean occursCheck() {
		if (this.hasVariableOnLeftHandSide() == false) {
			return false;
		}
		String variableName = this.getLeftAsVariable().getName();
		return FormulaHelper.occursCheck(variableName, this.right);
	}

	/**
	 * Applies the given substitution to both sides of the equality.
	 * 
	 * @param sub
	 *            the substitution to apply
	 */
	public void applySubstitution(final Substitution sub) {
		this.left = sub.applyTo(this.left);
		this.right = sub.applyTo(this.right);
	}

	/**
	 * Transforms the equality into a substitution.<br>
	 * This is only possible if the left-hand side is a variable.
	 * 
	 * @return The substitution which substitutes the left-hand variable of this
	 *         equality with the formula on the right-hand side of this
	 *         equality.
	 * @throws EqualityToSubstitutionException
	 *             : Thrown if this equality does not have a single variable on
	 *             its left-hand side.
	 */
	public Substitution toSubstitution() throws EqualityToSubstitutionException {
		if (!this.hasVariableOnLeftHandSide()) {
			throw new EqualityToSubstitutionException(this);
		}
		Substitution substitution = new Substitution();
		substitution.addReplacement(this.getLeftAsVariable().getName(), this.right);
		return substitution;
	}

	@Override
	public String toString() {
		return "EQ[" + this.left.toString() + " = " + this.right.toString() + "]";
	}

}
