package at.cl.atp.foresolution.model.unification;

import java.util.Map.Entry;
import java.util.Set;

import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.enums.AtomType;
import at.cl.atp.foresolution.model.substitution.Substitution;
import at.cl.atp.foresolution.model.utils.FormulaHelper;

public class Unification {

	/**
	 * Creates the Most General Unifier (MGU), a substitution that makes the
	 * left and right formulas syntactically equal.
	 * 
	 * @param left
	 *            The left formula
	 * @param right
	 *            The right formula
	 * @return The substitution that - applied to left and right individually -
	 *         makes them syntactically equal, or <code>null</code> if no such
	 *         substitution exists.
	 */
	public static Substitution createMGU(final Formula left, final Formula right) {
		UnificationProblem problem = new UnificationProblem(left, right, false);
		return problem.solve();
	}

	/**
	 * Creates a substitution for the <i>left</i> formula only that makes the
	 * left formula equal to the right formula.<br>
	 * The essential difference to an MGU is that the MGU may re-assign
	 * variables in both formulas, where the substitution created here only
	 * re-assigns variables in the left formula.
	 * 
	 * @param left
	 *            The formula which should be transformed
	 * @param right
	 *            The transformation target
	 * @return A substitution that - applied to the left formula - makes it
	 *         syntactically equal to the right formula, or <code>null</code> if
	 *         there is no such substitution.
	 */
	public static Substitution createSubstitution(final Formula left, final Formula right) {
		UnificationProblem problem = new UnificationProblem(left, right, true);
		Substitution sigma = problem.solve();
		if (sigma == null) {
			return null;
		}
		Set<String> varsInRight = FormulaHelper.getVariables(right);
		// if the substitution affects a variable from the right formula,
		// it must map the variable to itself, otherwise this is no
		// valid substitution for our purposes.
		for (Entry<String, Formula> mapping : sigma.getReplacements().entrySet()) {
			if (varsInRight.contains(mapping.getKey())) {
				// the formula must be equal to the variable
				Formula formula = mapping.getValue();
				if (formula instanceof Atom == false || ((Atom) formula).getType() != AtomType.VARIABLE) {
					// variable is not mapped to itself
					return null;
				}
				Atom variable = (Atom) formula;
				if (variable.contentEquals(mapping.getValue()) == false) {
					// variable is mapped to other variable, but not to
					// itself
					return null;
				}
			}
		}
		// mapping is okay
		return sigma;
	}

	public static boolean existsMGU(final Formula left, final Formula right) {
		Substitution mgu = createMGU(left, right);
		return mgu != null;
	}
}
