package at.cl.atp.foresolution.model.utils;

import java.util.HashSet;
import java.util.Set;

import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.BinaryFormula;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.Functor;
import at.cl.atp.foresolution.model.NegatedFormula;
import at.cl.atp.foresolution.model.QuantifiedFormula;
import at.cl.atp.foresolution.model.enums.AtomType;
import at.cl.atp.foresolution.model.enums.FunctorType;

public class FormulaHelper {
	/**
	 * Returns all free variables that occur in the formula.<br>
	 * A variable is considered to be <i>free</i> if there is no quantifier in
	 * the formula that binds the variable to it.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of all bound variables within the formula.
	 */
	public static Set<String> getFreeVariables(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		Set<String> boundVariables = new HashSet<String>();
		getFreeVariablesRec(formula, resultSet, boundVariables);
		return resultSet;
	}

	private static void getFreeVariablesRec(final Formula formula, final Set<String> freeVariables,
			final Set<String> boundInScope) {
		if (formula instanceof Atom) {
			Atom atom = (Atom) formula;
			if (atom.getType() == AtomType.VARIABLE) {
				if (!boundInScope.contains(atom.getName())) {
					freeVariables.add(atom.getName());
				}
			}
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			Set<String> newScope = new HashSet<String>(boundInScope);
			newScope.add(f.getQuantifiedVariable());
			getFreeVariablesRec(f.getInnerFormula(), freeVariables, newScope);
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			getFreeVariablesRec(f.getLeft(), freeVariables, boundInScope);
			getFreeVariablesRec(f.getRight(), freeVariables, boundInScope);
		} else if (formula instanceof NegatedFormula) {
			getFreeVariablesRec(((NegatedFormula) formula).getInnerFormula(), freeVariables, boundInScope);
		} else if (formula instanceof Functor) {
			for (Formula argument : ((Functor) formula).getArguments()) {
				getFreeVariablesRec(argument, freeVariables, boundInScope);
			}
		}
	}

	/**
	 * Returns all bound variables within the formula.<br>
	 * A variable is considered to be <i>bound</i> if there exists a quantifier
	 * in the formula that binds that variable to itself.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of bound variables within the formula
	 */
	public static Set<String> getBoundVariables(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getBoundVariablesRec(formula, resultSet);
		return resultSet;
	}

	private static void getBoundVariablesRec(final Formula formula, final Set<String> resultSet) {
		// note: we are only interested in the quantors here.
		// since they cannot appear inside functions or
		// predicates, we can safely ignore those constructs
		// if we encounter them. Since variables are bound
		// by quantors only (not by themselves) we can ignore
		// them as well.
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			resultSet.add(f.getQuantifiedVariable());
			getBoundVariablesRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			getBoundVariablesRec(f.getLeft(), resultSet);
			getBoundVariablesRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			getBoundVariablesRec(((NegatedFormula) formula).getInnerFormula(), resultSet);
		}
	}

	/**
	 * Gets all variables used in the formula.<br>
	 * Formally, this returns the <i>union</i> of all free variables and all
	 * bound variables in the formula.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of all variables occurring in the formula.
	 */
	public static Set<String> getVariables(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getVariablesRec(formula, resultSet);
		return resultSet;
	}

	private static void getVariablesRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof Atom) {
			Atom atom = (Atom) formula;
			if (atom.getType() == AtomType.VARIABLE) {
				resultSet.add(atom.getName());
			}
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getVariablesRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			getVariablesRec(f.getLeft(), resultSet);
			getVariablesRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			getVariablesRec(((NegatedFormula) formula).getInnerFormula(), resultSet);
		} else if (formula instanceof Functor) {
			for (Formula argument : ((Functor) formula).getArguments()) {
				getVariablesRec(argument, resultSet);
			}
		}
	}

	/**
	 * Returns the set of function names contained in the formula.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of function names occurring in the formula.
	 */
	public static Set<String> getFunctionSymbols(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getFunctionSymbolsRec(formula, resultSet);
		return resultSet;
	}

	private static void getFunctionSymbolsRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getFunctionSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			getFunctionSymbolsRec(f.getLeft(), resultSet);
			getFunctionSymbolsRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			getFunctionSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof Functor) {
			Functor f = (Functor) formula;
			if (f.getType() == FunctorType.FUNCTION) {
				resultSet.add(f.getName());
			}
			for (Formula argument : f.getArguments()) {
				getFunctionSymbolsRec(argument, resultSet);
			}
		}
	}

	/**
	 * Returns the set of predicate names contained in the formula.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of predicate names occurring in the formula.
	 */
	public static Set<String> getPredicateSymbols(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getPredicateSymbolsRec(formula, resultSet);
		return resultSet;
	}

	private static void getPredicateSymbolsRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getPredicateSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			getPredicateSymbolsRec(f.getLeft(), resultSet);
			getPredicateSymbolsRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			getPredicateSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof Functor) {
			Functor f = (Functor) formula;
			if (f.getType() == FunctorType.PREDICATE) {
				resultSet.add(f.getName());
			}
		}
	}

	/**
	 * Checks if the given variable occurs in the formula or not.
	 * 
	 * @param variable
	 *            The variable to search for
	 * @param formula
	 *            The formula to check
	 * @return <code>true</code> if the variable occurs in the formula,
	 *         otherwise <code>false</code>.
	 */
	public static boolean occursCheck(final String variable, final Formula formula) {
		Set<String> variables = getVariables(formula);
		return variables.contains(variable);
	}

	/**
	 * Checks if the given formula is a sentence (i.e. has no free variables).
	 * 
	 * @param formula
	 *            The formula to check
	 * @return <code>true</code> if formula is sentence, otherwise
	 *         <code>false</code>.
	 */
	public static boolean isSentence(final Formula formula) {
		return getFreeVariables(formula).isEmpty();
	}

	/**
	 * Returns the set of constant symbols in the formula.
	 * 
	 * @param formula
	 *            The formula to analyze
	 * @return The set of constants appearing within the formula.
	 */
	public static Set<String> getConstantSymbols(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getConstantSymbolsRec(formula, resultSet);
		return resultSet;
	}

	private static void getConstantSymbolsRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof Atom) {
			Atom atom = (Atom) formula;
			if (atom.getType() == AtomType.CONSTANT) {
				resultSet.add(atom.getName());
			}
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getConstantSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			getConstantSymbolsRec(f.getLeft(), resultSet);
			getConstantSymbolsRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			getConstantSymbolsRec(((NegatedFormula) formula).getInnerFormula(), resultSet);
		} else if (formula instanceof Functor) {
			for (Formula argument : ((Functor) formula).getArguments()) {
				getConstantSymbolsRec(argument, resultSet);
			}
		}
	}

	/**
	 * Counts all variable occurrences within the formula.<br>
	 * If the same variable appears multiple times within the formula, each
	 * occurrence will be counted. Variable bindings that occur inside a
	 * quantifier will not be counted, only variable references are of interest.
	 * 
	 * @param formula
	 *            The formula to analyze
	 * @return The total number of variable occurrences within the formula.
	 */
	public static int getVariableOccurrenceCount(final Formula formula) {
		if (formula instanceof Atom) {
			if (((Atom) formula).getType() == AtomType.VARIABLE) {
				return 1;
			}
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			return getVariableOccurrenceCount(f.getInnerFormula());
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			int temp = getVariableOccurrenceCount(f.getLeft());
			temp += getVariableOccurrenceCount(f.getRight());
			return temp;
		} else if (formula instanceof NegatedFormula) {
			return getVariableOccurrenceCount(((NegatedFormula) formula).getInnerFormula());
		} else if (formula instanceof Functor) {
			int temp = 0;
			for (Formula argument : ((Functor) formula).getArguments()) {
				temp += getVariableOccurrenceCount(argument);
			}
			return temp;
		}
		return 0;
	}

}
