package at.cl.atp.foresolution.transformations;

import java.util.HashSet;
import java.util.Set;

import at.cl.atp.foresolution.model.BinaryFormula;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.QuantifiedFormula;
import at.cl.atp.foresolution.model.TermRegistry;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.model.enums.BinaryOperation;

public class ClausalFormTransformation implements FormulaTransformation {

	public static final ClausalFormTransformation instance;

	static {
		instance = new ClausalFormTransformation();
	}

	private ClausalFormTransformation() {
	}

	private boolean changed = false;

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		Set<TransformationType> requirements = new HashSet<TransformationType>();
		requirements.add(TransformationType.RECTIFICATION);
		requirements.add(TransformationType.NEGATION_NORMAL_FORM);
		requirements.add(TransformationType.STRUCTURAL_SKOLEMIZATION);
		return requirements;
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.CLAUSAL_FORM)) {
			return;
		}
		TransformationProvider.resolvePreconditions(this, formula);
		this.changed = true;
		while (this.changed) {
			this.changed = false;
			formula.setFormula(this.transformRec(formula.getFormula()));
		}
	}

	private Formula transformRec(final Formula formula) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			return this.transformRec(f.getInnerFormula());
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			if (f.getOperation() == BinaryOperation.OR) {
				if (this.isConjunction(f.getLeft())) {
					// case (P and Q) or R --> (P or R) and (Q or R)
					BinaryFormula leftConjunction = (BinaryFormula) f.getLeft();
					Formula p = leftConjunction.getLeft();
					Formula q = leftConjunction.getRight();
					Formula r = f.getRight();
					BinaryFormula leftOr = TermRegistry.createDisjunction(p, r);
					BinaryFormula rightOr = TermRegistry.createDisjunction(q, r);
					BinaryFormula newAnd = TermRegistry.createConjunction(leftOr, rightOr);
					this.changed = true;
					return newAnd;
				} else if (this.isConjunction(f.getRight())) {
					// case P or (Q and R) --> (P or Q) and (P or R)
					BinaryFormula rightConjunction = (BinaryFormula) f.getRight();
					Formula p = f.getLeft();
					Formula q = rightConjunction.getLeft();
					Formula r = rightConjunction.getRight();
					BinaryFormula leftOr = TermRegistry.createDisjunction(p, q);
					BinaryFormula rightOr = TermRegistry.createDisjunction(p, r);
					BinaryFormula newAnd = TermRegistry.createConjunction(leftOr, rightOr);
					this.changed = true;
					return newAnd;
				} else {
					Formula left = this.transformRec(f.getLeft());
					Formula right = this.transformRec(f.getRight());
					return TermRegistry.createDisjunction(left, right);
				}
			} else {
				Formula left = this.transformRec(f.getLeft());
				Formula right = this.transformRec(f.getRight());
				return TermRegistry.createConjunction(left, right);
			}
		}
		return formula;
	}

	private boolean isConjunction(final Formula formula) {
		if (formula instanceof BinaryFormula) {
			return ((BinaryFormula) formula).getOperation() == BinaryOperation.AND;
		}
		return false;
	}

}
