package at.cl.atp.foresolution.transformations;

import java.util.Collections;
import java.util.Set;

import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.BinaryFormula;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.Functor;
import at.cl.atp.foresolution.model.NegatedFormula;
import at.cl.atp.foresolution.model.QuantifiedFormula;
import at.cl.atp.foresolution.model.TermRegistry;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.model.enums.BinaryOperation;
import at.cl.atp.foresolution.model.enums.Quantifier;

public class NNFTransformation implements FormulaTransformation {

	public static final NNFTransformation instance;

	static {
		instance = new NNFTransformation();
	}

	private NNFTransformation() {

	}

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		return Collections.emptySet();
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.NEGATION_NORMAL_FORM)) {
			return;
		}
		TransformationProvider.resolvePreconditions(this, formula);
		formula.setFormula(this.doTransform(formula.getFormula()));
		formula.getAppliedTransformations().add(TransformationType.NEGATION_NORMAL_FORM);
	}

	private Formula doTransform(final Formula formula) {
		if (formula instanceof NegatedFormula) {
			return this.pushNegationInside(((NegatedFormula) formula).getInnerFormula());
		} else {
			if (formula instanceof BinaryFormula) {
				BinaryFormula f = (BinaryFormula) formula;
				Formula left = this.doTransform(f.getLeft());
				Formula right = this.doTransform(f.getRight());
				if (f.getOperation() == BinaryOperation.AND) {
					return TermRegistry.createConjunction(left, right);
				} else {
					return TermRegistry.createDisjunction(left, right);
				}
			} else if (formula instanceof QuantifiedFormula) {
				QuantifiedFormula f = (QuantifiedFormula) formula;
				if (f.getQuantifier() == Quantifier.EXISTENTIAL) {
					return TermRegistry.createExistentialFormula(this.doTransform(f.getInnerFormula()),
							f.getQuantifiedVariable());
				} else {
					return TermRegistry.createUniversalFormula(this.doTransform(f.getInnerFormula()),
							f.getQuantifiedVariable());
				}
			}
			return formula;
		}
	}

	private Formula pushNegationInside(final Formula formula) {
		if (formula instanceof NegatedFormula) {
			// we have a double negation
			return this.doTransform(((NegatedFormula) formula).getInnerFormula());
		} else {
			if (formula instanceof BinaryFormula) {
				BinaryFormula f = (BinaryFormula) formula;
				Formula left = this.pushNegationInside(f.getLeft());
				Formula right = this.pushNegationInside(f.getRight());
				// we need to flip the operator around
				if (f.getOperation() == BinaryOperation.AND) {
					return TermRegistry.createDisjunction(left, right);
				} else {
					return TermRegistry.createConjunction(left, right);
				}
			} else if (formula instanceof QuantifiedFormula) {
				QuantifiedFormula f = (QuantifiedFormula) formula;
				// we need to swap the quantifier symbol
				Formula innerFormula = this.pushNegationInside(f.getInnerFormula());
				if (f.getQuantifier() == Quantifier.EXISTENTIAL) {
					return TermRegistry.createUniversalFormula(innerFormula, f.getQuantifiedVariable());
				} else {
					return TermRegistry.createExistentialFormula(innerFormula, f.getQuantifiedVariable());
				}
			} else if (formula instanceof Functor) {
				// negate the predicate
				return TermRegistry.createNegatedFormula(formula);
			} else if (formula instanceof Atom) {
				// negate the atom
				return TermRegistry.createNegatedFormula(formula);
			}
			return formula;
		}
	}

}
