package at.cl.atp.foresolution.transformations;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.BinaryFormula;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.Functor;
import at.cl.atp.foresolution.model.NegatedFormula;
import at.cl.atp.foresolution.model.QuantifiedFormula;
import at.cl.atp.foresolution.model.TermRegistry;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.model.enums.BinaryOperation;
import at.cl.atp.foresolution.model.enums.FunctorType;
import at.cl.atp.foresolution.model.enums.Quantifier;
import at.cl.atp.foresolution.model.substitution.Substitution;
import at.cl.atp.foresolution.model.utils.FormulaHelper;
import at.cl.atp.foresolution.settings.Settings;

public class RectificationTransformation implements FormulaTransformation {

	public static final RectificationTransformation instance;

	static {
		instance = new RectificationTransformation();
	}

	private RectificationTransformation() {
	}

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		return Collections.emptySet();
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.RECTIFICATION)) {
			return;
		}
		if (Settings.noRenaming() == false) {
			this.renameFunctionsAndPredicates(formula, formula.getFormula());
			Formula newFormula = this.renameVariablesRec(formula, formula.getFormula());
			formula.setFormula(newFormula);
		}
		formula.getAppliedTransformations().add(TransformationType.RECTIFICATION);
	}

	private void renameFunctionsAndPredicates(final FormulaContainer container, final Formula formula) {
		Set<String> functionNames = FormulaHelper.getFunctionSymbols(formula);
		Set<String> predicateNames = FormulaHelper.getPredicateSymbols(formula);
		Map<String, String> oldNameToNewName = new HashMap<String, String>();
		// create a new name and map the old name to it
		// note: we can use the same map for function and predicate names
		// here because the parser requires function names to start with
		// a lowercase letter while predicates must begin with an uppercase
		// letter, therefore we cannot have "collisions" between function
		// names and predicate names.
		for (String oldName : functionNames) {
			String newName = container.nextSkolemFunction();
			oldNameToNewName.put(oldName, newName);
		}
		for (String oldName : predicateNames) {
			String newName = container.nextSkolemPredicate();
			oldNameToNewName.put(oldName, newName);
		}
		// recursively walk through the tree, replacing function and predicate
		// names on the fly
		Formula newFormula = this.renameFunctionsAndPredicatesRec(formula, oldNameToNewName);
		container.setFormula(newFormula);
	}

	private Formula renameFunctionsAndPredicatesRec(final Formula formula, final Map<String, String> oldNameToNewName) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			if (f.getQuantifier() == Quantifier.EXISTENTIAL) {
				return TermRegistry.createExistentialFormula(
						this.renameFunctionsAndPredicatesRec(f.getInnerFormula(), oldNameToNewName),
						f.getQuantifiedVariable());
			} else {
				return TermRegistry.createUniversalFormula(
						this.renameFunctionsAndPredicatesRec(f.getInnerFormula(), oldNameToNewName),
						f.getQuantifiedVariable());
			}

		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			Formula left = this.renameFunctionsAndPredicatesRec(f.getLeft(), oldNameToNewName);
			Formula right = this.renameFunctionsAndPredicatesRec(f.getRight(), oldNameToNewName);
			if (f.getOperation() == BinaryOperation.AND) {
				return TermRegistry.createConjunction(left, right);
			} else {
				return TermRegistry.createDisjunction(left, right);
			}
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			return TermRegistry.createNegatedFormula(this.renameFunctionsAndPredicatesRec(f.getInnerFormula(),
					oldNameToNewName));
		} else if (formula instanceof Functor) {
			Functor f = (Functor) formula;
			String newName = oldNameToNewName.get(f.getName());
			Formula[] newArgs = new Formula[f.getArguments().length];
			for (int i = 0; i < newArgs.length; i++) {
				newArgs[i] = this.renameFunctionsAndPredicatesRec(f.getArguments()[i], oldNameToNewName);
			}
			if (f.getType() == FunctorType.FUNCTION) {
				return TermRegistry.createFunction(newName, newArgs);
			} else {
				return TermRegistry.createPredicate(newName, newArgs);
			}
		}
		return formula;
	}

	private Formula renameVariablesRec(final FormulaContainer container, final Formula formula) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			// map each variable to a new identifier
			String newVariableName = container.nextSkolemVariable();
			// perform a substitution of the variable names
			// in the scope of the quantifier
			Substitution substitution = new Substitution();
			// prepare a variable instance
			Atom variable = TermRegistry.createVariable(newVariableName);
			// add the replacement of { oldName -> newName }
			substitution.addReplacement(f.getQuantifiedVariable(), variable);
			// perform the substitution on the inner formula of the quantifier
			Formula newInner = substitution.applyTo(f.getInnerFormula());
			newInner = this.renameVariablesRec(container, newInner);
			if (f.getQuantifier() == Quantifier.EXISTENTIAL) {
				return TermRegistry.createExistentialFormula(newInner, newVariableName);
			} else {
				return TermRegistry.createUniversalFormula(newInner, newVariableName);
			}
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			// apply the transformation to both arguments
			Formula left = this.renameVariablesRec(container, f.getLeft());
			Formula right = this.renameVariablesRec(container, f.getRight());
			if (f.getOperation() == BinaryOperation.AND) {
				return TermRegistry.createConjunction(left, right);
			} else {
				return TermRegistry.createDisjunction(left, right);
			}
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			// apply the transformation to the inner formula
			Formula newInner = this.renameVariablesRec(container, f.getInnerFormula());
			return TermRegistry.createNegatedFormula(newInner);
		} else if (formula instanceof Functor) {
			Functor f = (Functor) formula;
			// apply the transformation to all arguments
			Formula[] newArgs = new Formula[f.getArguments().length];
			for (int i = 0; i < newArgs.length; i++) {
				newArgs[i] = this.renameVariablesRec(container, f.getArguments()[i]);
			}
			if (f.getType() == FunctorType.FUNCTION) {
				return TermRegistry.createFunction(f.getName(), newArgs);
			} else {
				return TermRegistry.createPredicate(f.getName(), newArgs);
			}
		}
		return formula;
	}
}
