package at.cl.atp.foresolution.transformations;

import java.util.HashSet;
import java.util.Set;

import at.cl.atp.foresolution.model.Atom;
import at.cl.atp.foresolution.model.BinaryFormula;
import at.cl.atp.foresolution.model.Formula;
import at.cl.atp.foresolution.model.Functor;
import at.cl.atp.foresolution.model.NegatedFormula;
import at.cl.atp.foresolution.model.QuantifiedFormula;
import at.cl.atp.foresolution.model.TermRegistry;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.model.enums.BinaryOperation;
import at.cl.atp.foresolution.model.enums.Quantifier;
import at.cl.atp.foresolution.model.substitution.Substitution;

public class StructuralSkolemization implements FormulaTransformation {

	public static final StructuralSkolemization instance;

	static {
		instance = new StructuralSkolemization();
	}

	private StructuralSkolemization() {
	}

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		Set<TransformationType> preconditions = new HashSet<TransformationType>();
		preconditions.add(TransformationType.RECTIFICATION);
		preconditions.add(TransformationType.NEGATION_NORMAL_FORM);
		return preconditions;
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.STRUCTURAL_SKOLEMIZATION)) {
			return;
		}
		TransformationProvider.resolvePreconditions(this, formula);
		formula.setFormula(this.skolemizeRec(formula, formula.getFormula(), Quantifier.EXISTENTIAL,
				new HashSet<String>()));
		formula.getAppliedTransformations().add(TransformationType.STRUCTURAL_SKOLEMIZATION);
	}

	private Formula skolemizeRec(final FormulaContainer container, final Formula formula,
			final Quantifier strongQuantifier, final Set<String> boundVariables) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			if (f.getQuantifier() == strongQuantifier) {
				// prepare a substitution for the variables bound by the strong
				// quantifier
				// to new skolem functions that depend on the bound variables
				Substitution substitution = new Substitution();
				String newFunctionName = container.nextSkolemFunction();
				// that function depends on all bound variables in our scope
				Formula[] newArgs = new Formula[boundVariables.size()];
				int index = 0;
				for (String boundVar : boundVariables) {
					Atom newVar = TermRegistry.createVariable(boundVar);
					newArgs[index] = newVar;
					index++;
				}
				Functor skolemFunction = TermRegistry.createFunction(newFunctionName, newArgs);
				substitution.addReplacement(f.getQuantifiedVariable(), skolemFunction);
				// apply the substitution to the sub-formula of the quantifier
				Formula newInner = substitution.applyTo(f.getInnerFormula());
				// note that we return the substituted inner formula, NOT the
				// quantifier!
				// this way we remove the quantifier from the formula.
				return this.skolemizeRec(container, newInner, strongQuantifier, boundVariables);
			} else {
				// add all variables that are bound
				Set<String> newBoundVariables = new HashSet<String>(boundVariables);
				newBoundVariables.add(f.getQuantifiedVariable());
				if (f.getQuantifier() == Quantifier.EXISTENTIAL) {
					return TermRegistry.createExistentialFormula(
							this.skolemizeRec(container, f.getInnerFormula(), strongQuantifier, newBoundVariables),
							f.getQuantifiedVariable());
				} else {
					return TermRegistry.createUniversalFormula(
							this.skolemizeRec(container, f.getInnerFormula(), strongQuantifier, newBoundVariables),
							f.getQuantifiedVariable());
				}
			}
		} else if (formula instanceof BinaryFormula) {
			BinaryFormula f = (BinaryFormula) formula;
			Formula left = this.skolemizeRec(container, f.getLeft(), strongQuantifier, boundVariables);
			Formula right = this.skolemizeRec(container, f.getRight(), strongQuantifier, boundVariables);
			if (f.getOperation() == BinaryOperation.AND) {
				return TermRegistry.createConjunction(left, right);
			} else {
				return TermRegistry.createDisjunction(left, right);
			}
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			// swap around the strong quantifier and apply to inner formula
			if (strongQuantifier == Quantifier.EXISTENTIAL) {
				return TermRegistry.createNegatedFormula(this.skolemizeRec(container, f.getInnerFormula(),
						Quantifier.UNIVERSAL, boundVariables));
			} else {
				return TermRegistry.createNegatedFormula(this.skolemizeRec(container, f.getInnerFormula(),
						Quantifier.EXISTENTIAL, boundVariables));
			}
		}
		// in all other cases (predicates, functions, atoms) we leave
		// the formula unchanged
		return formula;
	}
}
