package at.cl.atp.foresolution.alphabet;

import java.util.HashMap;
import java.util.Map;

public class GreekAlphabet implements Alphabet {

	private static final int ALPHABET_SIZE = 24;

	private static final char[] LETTERS_LOWER_UNICODE = { '\u03B1', '\u03B2', '\u03b3', '\u03b4', '\u03b5', '\u03b6',
			'\u03b7', '\u03b8', '\u03b9', '\u03ba', '\u03bb', '\u03bc', '\u03bd', '\u03be', '\u03bf', '\u03c0',
			'\u03c1', '\u03c2', '\u03c3', '\u03c4', '\u03c5', '\u03c6', '\u03c7', '\u03c8', '\u03c9' };

	private static final char[] LETTERS_UPPER_UNICODE = { '\u0391', '\u0392', '\u0393', '\u0394', '\u0395', '\u0396',
			'\u0397', '\u0398', '\u0399', '\u039a', '\u039b', '\u039c', '\u039d', '\u039e', '\u039f', '\u03a0',
			'\u03a1', '\u03a2', '\u03a3', '\u03a4', '\u03a5', '\u03a6', '\u03a7', '\u03a8', '\u03a9' };

	private static final String[] LETTERS_LOWER_LATEX = { "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta",
			"theta", "iota", "kappa", "lambda", "mu", "nu", "xi", "omicron", "pi", "rho", "sigma", "tau", "upsilon",
			"phi", "chi", "psi", "omega" };

	private static final String[] LETTERS_UPPER_LATEX = { "Alpha", "Beta", "Gamma", "Delta", "Epsilon", "Zeta", "Eta",
			"Theta", "Iota", "Kappa", "Lambda", "Mu", "Nu", "Xi", "Omicron", "Pi", "Rho", "Sigma", "Tau", "Upsilon",
			"Phi", "Chi", "Psi", "Omega" };

	// replace by array for efficiency
	private static final Map<Character, Integer> LOWER_UNICODE_LETTER_INDEX;
	private static final Map<Character, Integer> UPPER_UNICODE_LETTER_INDEX;
	private static final Map<String, Integer> LOWER_LATEX_LETTER_INDEX;
	private static final Map<String, Integer> UPPER_LATEX_LETTER_INDEX;

	public static final Alphabet instance;

	static {
		LOWER_UNICODE_LETTER_INDEX = new HashMap<Character, Integer>();
		UPPER_UNICODE_LETTER_INDEX = new HashMap<Character, Integer>();
		LOWER_LATEX_LETTER_INDEX = new HashMap<String, Integer>();
		UPPER_LATEX_LETTER_INDEX = new HashMap<String, Integer>();
		for (int i = 0; i < ALPHABET_SIZE; i++) {
			LOWER_UNICODE_LETTER_INDEX.put(LETTERS_LOWER_UNICODE[i], i);
			UPPER_UNICODE_LETTER_INDEX.put(LETTERS_UPPER_UNICODE[i], i);
			LOWER_LATEX_LETTER_INDEX.put(LETTERS_LOWER_LATEX[i], i);
			UPPER_LATEX_LETTER_INDEX.put(LETTERS_UPPER_LATEX[i], i);
		}
		instance = new GreekAlphabet();
	}

	private GreekAlphabet() {
	}

	private boolean isIndexInRange(final int index) {
		if (index < 0 || index > ALPHABET_SIZE) {
			return false;
		} else {
			return true;
		}
	}

	private void rangeCheck(final int index) {
		if (!this.isIndexInRange(index)) {
			throw new RuntimeException("Unmapped index " + index + " for greek alphabet!");
		}
	}

	@Override
	public char getUnicodeLower(final int characterIndex) {
		this.rangeCheck(characterIndex);
		return LETTERS_LOWER_UNICODE[characterIndex];
	}

	@Override
	public char getUnicodeUpper(final int characterIndex) {
		this.rangeCheck(characterIndex);
		return LETTERS_UPPER_UNICODE[characterIndex];
	}

	@Override
	public String getLatexLower(final int characterIndex, final boolean includeBackslash) {
		this.rangeCheck(characterIndex);
		String character = LETTERS_LOWER_LATEX[characterIndex];
		if (includeBackslash) {
			character = "\\" + character;
		}
		return character;
	}

	@Override
	public String getLatexUpper(final int characterIndex, final boolean includeBackslash) {
		this.rangeCheck(characterIndex);
		String character = LETTERS_UPPER_LATEX[characterIndex];
		if (includeBackslash) {
			character = "\\" + character;
		}
		return character;
	}

	@Override
	public int getIndex(final char unicodeCharacter) {
		if (LOWER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)) {
			return LOWER_UNICODE_LETTER_INDEX.get(unicodeCharacter);
		} else if (UPPER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)) {
			return UPPER_UNICODE_LETTER_INDEX.get(unicodeCharacter);
		} else {
			throw new RuntimeException("Character " + unicodeCharacter + " is no greek letter!");
		}
	}

	@Override
	public int getIndex(final String latexCharacter) {
		String actualCharacter = latexCharacter.replaceAll("\\\\", "").trim();
		if (LOWER_LATEX_LETTER_INDEX.containsKey(actualCharacter)) {
			return LOWER_LATEX_LETTER_INDEX.get(actualCharacter);
		} else if (UPPER_LATEX_LETTER_INDEX.containsKey(actualCharacter)) {
			return UPPER_LATEX_LETTER_INDEX.get(actualCharacter);
		} else {
			throw new RuntimeException("Character " + latexCharacter + " is no greek letter!");
		}
	}

	@Override
	public char toUpper(final char unicodeCharacter) {
		int index = this.getIndex(unicodeCharacter);
		return LETTERS_UPPER_UNICODE[index];
	}

	@Override
	public char toLower(final char unicodeCharacter) {
		int index = this.getIndex(unicodeCharacter);
		return LETTERS_LOWER_UNICODE[index];
	}

	@Override
	public String toUpper(final String latexCharacter) {
		int index = this.getIndex(latexCharacter);
		return LETTERS_UPPER_LATEX[index];
	}

	@Override
	public String toLower(final String latexCharacter) {
		int index = this.getIndex(latexCharacter);
		return LETTERS_LOWER_LATEX[index];
	}

	@Override
	public int getAlphabetSize() {
		return ALPHABET_SIZE;
	}

	@Override
	public String toLatex(final char unicodeCharacter) {
		if (LOWER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)) {
			int index = LOWER_UNICODE_LETTER_INDEX.get(unicodeCharacter);
			return LETTERS_LOWER_LATEX[index];
		} else if (UPPER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)) {
			int index = UPPER_UNICODE_LETTER_INDEX.get(unicodeCharacter);
			return LETTERS_UPPER_LATEX[index];
		} else {
			throw new RuntimeException("Character " + unicodeCharacter + " is no greek letter!");
		}
	}

	@Override
	public char toUnicode(final String latexCharacter) {
		String actualCharacter = latexCharacter.replaceAll("\\\\", "").trim();
		if (LOWER_LATEX_LETTER_INDEX.containsKey(actualCharacter)) {
			int index = LOWER_LATEX_LETTER_INDEX.get(actualCharacter);
			return LETTERS_LOWER_UNICODE[index];
		} else if (UPPER_LATEX_LETTER_INDEX.containsKey(actualCharacter)) {
			int index = UPPER_LATEX_LETTER_INDEX.get(actualCharacter);
			return LETTERS_UPPER_UNICODE[index];
		} else {
			throw new RuntimeException("Character " + latexCharacter + " is no greek letter!");
		}
	}

	@Override
	public boolean contains(final char unicodeCharacter) {
		return LOWER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)
				|| UPPER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter);
	}

	@Override
	public boolean isUppercase(final char unicodeCharacter) {
		return UPPER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter);
	}

	@Override
	public boolean isLowercase(final char unicodeCharacter) {
		return LOWER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter);
	}
}
