package at.cl.atp.foresolution.alphabet;

import java.util.HashMap;
import java.util.Map;

public class LatinAlphabet implements Alphabet {

	private static final int ALPHABET_SIZE = 24;

	private static final char[] LETTERS_LOWER_UNICODE = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
			'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z' };
	private static final char[] LETTERS_UPPER_UNICODE = { 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
			'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z' };

	private static final Map<Character, Integer> LOWER_UNICODE_LETTER_INDEX;
	private static final Map<Character, Integer> UPPER_UNICODE_LETTER_INDEX;

	public static final LatinAlphabet instance;

	static {
		instance = new LatinAlphabet();
		LOWER_UNICODE_LETTER_INDEX = new HashMap<Character, Integer>();
		UPPER_UNICODE_LETTER_INDEX = new HashMap<Character, Integer>();
		for (int i = 0; i < ALPHABET_SIZE; i++) {
			LOWER_UNICODE_LETTER_INDEX.put(LETTERS_LOWER_UNICODE[i], i);
			UPPER_UNICODE_LETTER_INDEX.put(LETTERS_UPPER_UNICODE[i], i);
		}
	}

	private LatinAlphabet() {
	}

	private boolean isIndexInRange(final int index) {
		if (index < 0 || index > ALPHABET_SIZE) {
			return false;
		} else {
			return true;
		}
	}

	private void rangeCheck(final int index) {
		if (!this.isIndexInRange(index)) {
			throw new RuntimeException("Unmapped index " + index + " for greek alphabet!");
		}
	}

	@Override
	public char getUnicodeLower(final int characterIndex) {
		this.rangeCheck(characterIndex);
		return LETTERS_LOWER_UNICODE[characterIndex];
	}

	@Override
	public char getUnicodeUpper(final int characterIndex) {
		this.rangeCheck(characterIndex);
		return LETTERS_UPPER_UNICODE[characterIndex];
	}

	@Override
	public String getLatexLower(final int characterIndex, final boolean includeBackslash) {
		this.rangeCheck(characterIndex);
		return "" + LETTERS_LOWER_UNICODE[characterIndex];
	}

	@Override
	public String getLatexUpper(final int characterIndex, final boolean includeBackslash) {
		this.rangeCheck(characterIndex);
		return "" + LETTERS_UPPER_UNICODE[characterIndex];
	}

	@Override
	public int getIndex(final char unicodeCharacter) {
		if (LOWER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)) {
			return LOWER_UNICODE_LETTER_INDEX.get(unicodeCharacter);
		} else if (UPPER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter)) {
			return UPPER_UNICODE_LETTER_INDEX.get(unicodeCharacter);
		} else {
			throw new RuntimeException("Character " + unicodeCharacter + " is no greek letter!");
		}
	}

	@Override
	public int getIndex(final String latexCharacter) {
		String actualCharacter = latexCharacter.replaceAll("\\\\", "").trim();
		if (LOWER_UNICODE_LETTER_INDEX.containsKey(actualCharacter)) {
			return LOWER_UNICODE_LETTER_INDEX.get(actualCharacter);
		} else if (UPPER_UNICODE_LETTER_INDEX.containsKey(actualCharacter)) {
			return UPPER_UNICODE_LETTER_INDEX.get(actualCharacter);
		} else {
			throw new RuntimeException("Character " + latexCharacter + " is no greek letter!");
		}
	}

	@Override
	public char toUpper(final char unicodeCharacter) {
		int index = this.getIndex(unicodeCharacter);
		return LETTERS_UPPER_UNICODE[index];
	}

	@Override
	public char toLower(final char unicodeCharacter) {
		int index = this.getIndex(unicodeCharacter);
		return LETTERS_LOWER_UNICODE[index];
	}

	@Override
	public String toUpper(final String latexCharacter) {
		int index = this.getIndex(latexCharacter);
		return "" + LETTERS_UPPER_UNICODE[index];
	}

	@Override
	public String toLower(final String latexCharacter) {
		int index = this.getIndex(latexCharacter);
		return "" + LETTERS_LOWER_UNICODE[index];
	}

	@Override
	public int getAlphabetSize() {
		return ALPHABET_SIZE;
	}

	@Override
	public String toLatex(final char unicodeCharacter) {
		this.getIndex(unicodeCharacter);
		return "" + unicodeCharacter;
	}

	@Override
	public char toUnicode(final String latexCharacter) {
		this.getIndex(latexCharacter);
		return latexCharacter.charAt(0);
	}

	@Override
	public boolean contains(final char unicodeCharacter) {
		try {
			this.getIndex(unicodeCharacter);
			return true;
		} catch (RuntimeException e) {
			return false;
		}
	}

	@Override
	public boolean isUppercase(final char unicodeCharacter) {
		return UPPER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter);
	}

	@Override
	public boolean isLowercase(final char unicodeCharacter) {
		return LOWER_UNICODE_LETTER_INDEX.containsKey(unicodeCharacter);
	}

}
