package at.cl.atp.foresolution.main;

import java.io.File;

import org.eclipse.emf.ecore.util.EcoreUtil;

import at.cl.atp.fOResolution.FOResolutionFactory;
import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.foresolution.latexExporter.LatexExporter;
import at.cl.atp.foresolution.model.clause.ClausalForm;
import at.cl.atp.foresolution.model.clause.ClausalFormBuilder;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.parser.FirstOrderDSLParser;
import at.cl.atp.foresolution.prover.FirstOrderResolutionProver;
import at.cl.atp.foresolution.prover.ResolutionOutput;
import at.cl.atp.foresolution.prover.ResolutionResult;
import at.cl.atp.foresolution.settings.Settings;
import at.cl.atp.foresolution.transformations.NormalFormTransformations;
import at.cl.atp.foresolution.util.FormulaFormatter;
import at.cl.atp.foresolution.util.FormulaHelper;

public class FirstOrderResolutionProverMain {

	/**
	 * @param args
	 */
	public static void main(final String[] args) {
		long timeBefore = System.currentTimeMillis();
		File source = new File("test.fores");
		if (source.exists() == false) {
			System.err.println("Input file not found.");
			System.exit(-1);
		}
		Formula originalFormula = FirstOrderDSLParser.parseAndConvertToFormula(source);
		// take care that the formula is a sentence
		if (!FormulaHelper.isSentence(originalFormula)) {
			System.err.println("The formula is not a closed sentence!");
			System.exit(-1);
		}

		System.out.println("ORIGINAL FORMULA:");
		System.out.println("\t" + FormulaFormatter.format(originalFormula));
		System.out.println("\tFree Variables:  " + FormulaHelper.getFreeVariables(originalFormula).toString());
		System.out.println("\tBound Variables: " + FormulaHelper.getBoundVariables(originalFormula).toString());
		System.out.println("\tConstant names:  " + FormulaHelper.getConstantSymbols(originalFormula).toString());
		System.out.println("\tFunction names:  " + FormulaHelper.getFunctionSymbols(originalFormula).toString());
		System.out.println("\tPredicate names: " + FormulaHelper.getPredicateSymbols(originalFormula).toString());
		System.out.println();
		System.out.println("Negating formula...");
		Formula formula = null;
		if (Settings.negateInput()) {
			// negate the formula
			NegatedFormula negation = FOResolutionFactory.eINSTANCE.createNegatedFormula();
			negation.setInnerFormula(EcoreUtil.copy(originalFormula));
			formula = negation;
		} else {
			formula = EcoreUtil.copy(originalFormula);
		}
		// build a formula container for metadata
		FormulaContainer container = new FormulaContainer(formula);
		System.out.println("Rectified formula:");
		NormalFormTransformations.rectify(container);
		System.out.println("\t" + container.toString());
		System.out.println("Negation Normal Form:");
		NormalFormTransformations.toNNF(container);
		System.out.println("\t" + container.toString());
		System.out.println("After structural skolemization:");
		NormalFormTransformations.structuralSkolemization(container);
		System.out.println("\t" + container.toString());
		System.out.println("Clause Normal Form:");
		NormalFormTransformations.toClausalForm(container);
		System.out.println("\t" + container.toString());
		ClausalForm clausalForm = ClausalFormBuilder.build(container);
		System.out.println("Clausal Form:");
		System.out.println("\t" + clausalForm.toString());
		System.out.println("Running First Order Resolution Prover...");
		long beforeProver = System.currentTimeMillis();
		ResolutionOutput resolutionOutput = FirstOrderResolutionProver.run(container);
		long afterProver = System.currentTimeMillis();
		ResolutionResult result = resolutionOutput.getResult();
		LatexExporter.exportToLatex(resolutionOutput, originalFormula);
		long timeAfter = System.currentTimeMillis();
		System.out.println("RESULT IS: " + result);
		switch (result) {
		case BOX:
			System.out.println("\t--> The original formula is VALID.");
			break;
		case NO_CONTRADICTION:
			System.out.println("\t--> The original formula is SATISFIABLE but NOT VALID.");
			break;
		case EMPTY_SET:
			System.out.println("\t--> The original formula is UNSATISFIABLE.");
			break;
		}
		System.out.println();
		System.out.println();
		System.out.println("___TIME__________________________________________");
		System.out.println("Prover time: " + (afterProver - beforeProver) + "ms.");
		System.out.println("Total time: " + (timeAfter - timeBefore) + "ms.");
		long memTotal = Runtime.getRuntime().totalMemory();
		long memMax = Runtime.getRuntime().maxMemory();
		long memUsed = memTotal - Runtime.getRuntime().freeMemory();
		System.out.println();
		System.out.println();
		System.out.println("___MEMORY________________________________________");
		System.out.println("Available: " + toMegabytes(memMax));
		System.out.println("Allocated: " + toMegabytes(memTotal));
		System.out.println("Used:      " + toMegabytes(memUsed));
	}

	private static String toMegabytes(final long bytes) {
		double megabytes = bytes;
		megabytes /= 1024;
		megabytes /= 1024;
		return String.format("%.2f MB", megabytes);
	}
}
