package at.cl.atp.foresolution.model.clause;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import at.cl.atp.foresolution.exceptions.ClashProducedBoxException;
import at.cl.atp.foresolution.prover.ClashInfo;
import at.cl.atp.foresolution.subsumption.LiteralMatch;
import at.cl.atp.foresolution.subsumption.SubsumptionSearchTreeNode;
import at.cl.atp.foresolution.transformations.substitution.Substitution;
import at.cl.atp.foresolution.transformations.substitution.mgu.Unification;
import at.cl.atp.foresolution.util.SetOperations;

public class Clause {

	/**
	 * Creates a clause with a single literal.<br>
	 * The returned clause has all its parent and parentMGU references set to
	 * <code>null</code> and is intended for temporary use only.
	 * 
	 * @param literal
	 *            The literal to be contained in the unit clause
	 * @return The created unit clause containing the literal.
	 */
	public static Clause createUnitClause(final ClauseLiteral literal) {
		Clause clause = new Clause(null, null, null, Collections.singleton(literal));
		return clause;
	}

	private final Clause leftParent;
	private final Clause rightParent;
	private final Substitution parentMGU;

	private final Set<ClauseLiteral> literals;
	private final Set<ClauseLiteral> negatedLiterals;
	private final Set<Clause> clashedWith;

	private final int hashCode;
	private final String stringRepresentation;
	private final boolean isGround;

	public Clause(final Clause leftParent, final Clause rightParent, final Substitution parentMGU,
			final Set<ClauseLiteral> literals) {
		Set<ClauseLiteral> positiveLiterals = new HashSet<ClauseLiteral>();
		Set<ClauseLiteral> negativeLiterals = new HashSet<ClauseLiteral>();
		this.clashedWith = new HashSet<Clause>();
		this.leftParent = leftParent;
		this.rightParent = rightParent;
		this.parentMGU = parentMGU;
		for (ClauseLiteral literal : literals) {
			if (literal.isNegated()) {
				negativeLiterals.add(literal);
			} else {
				positiveLiterals.add(literal);
			}
		}
		this.literals = new ImmutableSet<ClauseLiteral>(positiveLiterals);
		this.negatedLiterals = new ImmutableSet<ClauseLiteral>(negativeLiterals);
		this.stringRepresentation = this.toStringInternal();
		this.hashCode = this.stringRepresentation.hashCode();
		int varCount = 0;
		for (ClauseLiteral literal : this.literals) {
			varCount += literal.getVariables().size();
		}
		for (ClauseLiteral literal : this.negatedLiterals) {
			varCount += literal.getVariables().size();
		}
		this.isGround = varCount <= 0;
	}

	public Set<ClauseLiteral> getLiterals() {
		return this.literals;
	}

	public Set<ClauseLiteral> getNegatedLiterals() {
		return this.negatedLiterals;
	}

	// private void removeLiteral(final ClauseLiteral literal) {
	// if (literal.isNegated()) {
	// this.negatedLiterals.remove(literal);
	// } else {
	// this.literals.remove(literal);
	// }
	// // this.changed = true;
	// }

	// public void addLiteral(final ClauseLiteral literal) {
	// if (literal.isNegated()) {
	// this.negatedLiterals.add(literal);
	// } else {
	// this.literals.add(literal);
	// }
	// // this.changed = true;
	// }

	public boolean isEmpty() {
		return this.literals.isEmpty() & this.negatedLiterals.isEmpty();
	}

	public boolean containsLiteralNegated(final ClauseLiteral literal) {
		return this.negatedLiterals.contains(literal);
	}

	public boolean containsLiteralNonNegated(final ClauseLiteral literal) {
		return this.literals.contains(literal);
	}

	public boolean containsLiteral(final ClauseLiteral literal) {
		return this.containsLiteralNegated(literal) || this.containsLiteralNonNegated(literal);
	}

	public boolean isTautology() {
		for (ClauseLiteral literal : this.literals) {
			for (ClauseLiteral negLiteral : this.negatedLiterals) {
				if (literal.getClashingMGU(negLiteral) != null) {
					return true;
				}
			}
		}
		return false;
	}

	public boolean hasClashedWith(final Clause other) {
		if (this.clashedWith.contains(other)) {
			return true;
		} else {
			return false;
		}
	}

/**
	 * Performs a <i>clash</i> operation with this and the other clause.<br>
	 * This operation will return a set of clauses (possibly empty) that
	 * contains clauses C that result from combining this clause with the other
	 * clause and removing a single pair of clashing literals from C.<br><br>
	 * 
	 * <b>NOTE</b>: The clash between the two clauses will be registered. Calling
	 * this operation multiple times on the same pair of clauses will always yield
	 * the empty set after the operation was executed for the first time. To clear
	 * this information, please use {@link Clause#clearClashCache()).
	 * 
	 * @param other The other clause to clash with
	 * @return The set of clauses resulting from the clash operation, or the empty
	 * set if this clause has already clashed with the given other clause.
	 * 
	 * @throws ClashProducedBoxException Thrown if a clash produced the empty clause. 
	 * This is a shortcut for performance reasons, because as soon as the empty clause is derived, 
	 * there is no need to derive any more clauses in the current iteration.
	 */
	public Set<Clause> clash(final Clause other) throws ClashProducedBoxException {
		if (this.clashedWith.contains(other)) {
			return Collections.emptySet();
		}
		this.clashedWith.add(other);
		other.clashedWith.add(this);
		Set<Clause> resultSet = new HashSet<Clause>();
		Set<ClashInfo> clashingLiterals = new HashSet<ClashInfo>();
		for (ClauseLiteral myLiteral : this.literals) {
			for (ClauseLiteral yourNegLiteral : other.negatedLiterals) {
				Substitution mgu = myLiteral.getClashingMGU(yourNegLiteral);
				if (mgu != null) {
					clashingLiterals.add(new ClashInfo(this, other, myLiteral, yourNegLiteral, mgu));
				}
			}
		}
		for (ClauseLiteral myNegLiteral : this.negatedLiterals) {
			for (ClauseLiteral yourLiteral : other.literals) {
				Substitution mgu = myNegLiteral.getClashingMGU(yourLiteral);
				if (mgu != null) {
					clashingLiterals.add(new ClashInfo(this, other, myNegLiteral, yourLiteral, mgu));
				}
			}
		}
		for (ClashInfo clashInfo : clashingLiterals) {
			Set<ClauseLiteral> literals = new HashSet<ClauseLiteral>();
			literals.addAll(this.literals);
			literals.addAll(other.literals);
			literals.addAll(this.negatedLiterals);
			literals.addAll(other.negatedLiterals);
			literals.remove(clashInfo.getFirstLiteral());
			literals.remove(clashInfo.getSecondLiteral());

			Clause resultingClause = new Clause(clashInfo.getFirstClause(), clashInfo.getSecondClause(),
					clashInfo.getMgu(), literals);
			if (resultingClause.isEmpty()) {
				// Shortcut for performance:
				// we've derived the empty clause
				throw new ClashProducedBoxException(clashInfo.getFirstClause(), clashInfo.getSecondClause());
			}
			// we only consider clauses that are not tautologies
			if (resultingClause.isTautology()) {
				continue;
			}
			if (resultingClause.isChangedBy(clashInfo.getMgu())) {
				// we now create a deep copy of that new clause and apply the
				// MGU (that made the literals clash) to the entire clause
				resultSet.add(resultingClause.applySubstitution(clashInfo.getMgu(), this, other));
			} else {
				// the MGU is empty or does not map a variable from the
				// remaining clause literals
				resultSet.add(resultingClause);
			}
		}
		return resultSet;
	}

	/**
	 * Determines if the given substitution will have any effect on this clause.<br>
	 * In other words, if this method returns <code>false</code>, then the
	 * clause will remain the same when applying the substitution on it.
	 * 
	 * @param sigma
	 *            The substitution to check
	 * @return <code>true</code> if the application of the substitution would
	 *         have an effect on this clause, otherwise <code>false</code>.
	 */
	private boolean isChangedBy(final Substitution mgu) {
		Set<String> mappedVariables = mgu.getReplacements().keySet();
		Set<String> clauseVariables = new HashSet<String>();
		for (ClauseLiteral literal : this.literals) {
			clauseVariables.addAll(literal.getVariables());
		}
		for (ClauseLiteral negatedLiteral : this.negatedLiterals) {
			clauseVariables.addAll(negatedLiteral.getVariables());
		}
		int sizeBefore = clauseVariables.size();
		clauseVariables.removeAll(mappedVariables);
		int sizeAfter = clauseVariables.size();
		if (sizeBefore == sizeAfter) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * Applies the given substitution to this clause.<br>
	 * This clause will not modified in the process.
	 * 
	 * @param sigma
	 *            The substitution to apply to this clause.
	 * @param leftParent
	 *            The left parent clause for the resulting clause
	 * @param rightParent
	 *            The right parent clause for the resulting clause
	 * @return The copy of this clause which has the substitution applied to it.
	 */
	private Clause applySubstitution(final Substitution sigma, final Clause leftParent, final Clause rightParent) {
		if (sigma.isEmpty()) {
			return this;
		}
		Set<ClauseLiteral> substitutedLiterals = new HashSet<ClauseLiteral>();
		for (ClauseLiteral literal : this.literals) {
			substitutedLiterals.add(literal.applySubstitution(sigma));
		}
		for (ClauseLiteral negatedLiteral : this.negatedLiterals) {
			substitutedLiterals.add(negatedLiteral.applySubstitution(sigma));
		}
		return new Clause(leftParent, rightParent, sigma, substitutedLiterals);
	}

	public void clearClashCache() {
		this.clashedWith.clear();
	}

	/**
	 * Checks if this formula subsumes the given 'other' formula.<br>
	 * This means that for every literal (L1) in this clause there must be at
	 * least one literal in the other clause (L2) where an MGU (sigma) exists
	 * such that L1sigma = L2sigma <b>and</b> each literal in the 'other'
	 * formula is used <i>at most once</i> for matching. Also, all individual
	 * MGUs may either not replace the same variables, or if they do, they must
	 * replace the same variable with the same term, resulting in a 'common' mgu
	 * that turns the set of literals in this clause into a multiset-subset of
	 * the literals in the other clause.
	 * 
	 * @param other
	 *            The other clause to check if this clause subsumes it
	 * @return <code>true</code> if <code>this</code> subsumes
	 *         <code>other</code>, otherwise <code>false</code>.
	 */
	public boolean subsumes(final Clause other) {
		if (this.literals.size() > other.literals.size()) {
			return false;
		}
		if (this.isGround()) {
			return false;
		}
		Set<LiteralMatch> matches = new HashSet<LiteralMatch>();
		Set<ClauseLiteral> unmatchedLiterals = new HashSet<ClauseLiteral>();
		Set<ClauseLiteral> myLiterals = new HashSet<ClauseLiteral>();
		myLiterals.addAll(this.literals);
		myLiterals.addAll(this.negatedLiterals);
		Set<ClauseLiteral> yourLiterals = new HashSet<ClauseLiteral>();
		yourLiterals.addAll(other.literals);
		yourLiterals.addAll(other.negatedLiterals);
		for (ClauseLiteral myLiteral : myLiterals) {
			boolean matched = false;
			for (ClauseLiteral yourLiteral : yourLiterals) {
				if (myLiteral.isSyntacticallyEqualTo(yourLiteral)) {
					// if we have equal literals, we do not consider
					// them any more for the subsumption problem.
					matched = true;
					break;
				}
				Substitution sigma = Unification.createSubstitution(myLiteral.getFormula(), yourLiteral.getFormula());
				if (sigma != null) {
					// found potential literal match
					LiteralMatch match = new LiteralMatch();
					match.setLeft(myLiteral);
					match.setRight(yourLiteral);
					match.setSigma(sigma);
					matches.add(match);
					unmatchedLiterals.add(myLiteral);
					matched = true;
				}
			}
			if (!matched) {
				// we did not find a match partner for the literal
				// so there cannot be a subsumption
				return false;
			}
		}
		if (matches.isEmpty()) {
			// the clauses were syntactically equal (modulo ordering of
			// literals)
			return true;
		}
		SubsumptionSearchTreeNode root = new SubsumptionSearchTreeNode();
		root.getUnmatchedLiterals().addAll(unmatchedLiterals);
		root.getMatches().addAll(matches);
		Substitution mgu = root.expandChildren();
		if (mgu != null) {
			// System.out.println("INFO :: Found subsumption. " +
			// this.toString() + " subsumes " + other.toString()
			// + " using the substitution " + mgu.toString());
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Checks if this clause is "ground", i.e. has no variables in any of its
	 * literals.
	 * 
	 * @return <code>true</code> if this clause is ground, otherwise
	 *         <code>false</code>.
	 */
	private boolean isGround() {
		return this.isGround;
	}

	/**
	 * Creates the string representation of this clause from scratch.
	 * 
	 * @return The newly computed string representation.
	 */
	private String toStringInternal() {
		StringBuilder builder = new StringBuilder();
		builder.append("{ ");
		Iterator<ClauseLiteral> iterator = this.literals.iterator();
		while (iterator.hasNext()) {
			ClauseLiteral literal = iterator.next();
			builder.append(literal.toString());
			if (iterator.hasNext() || !this.negatedLiterals.isEmpty()) {
				builder.append(", ");
			}
		}
		iterator = this.negatedLiterals.iterator();
		while (iterator.hasNext()) {
			ClauseLiteral literal = iterator.next();
			builder.append(literal.toString());
			if (iterator.hasNext()) {
				builder.append(", ");
			}
		}
		builder.append(" }");
		return builder.toString();
	}

	// /**
	// * Recalculates the hashCode of this Clause, along with the result of the
	// * {@link Clause#toString()} operation.<br>
	// * Does nothing if the clause has not changed since the last time this
	// * operation was called.
	// */
	// private void recalculateHashCodeAndStringRepresentation() {
	// this.stringRepresentation = this.toStringInternal();
	// this.hashCode = this.stringRepresentation.hashCode();
	// }

	@Override
	public int hashCode() {
		// this.recalculateHashCodeAndStringRepresentation();
		return this.hashCode;
	}

	@Override
	public String toString() {
		// this.recalculateHashCodeAndStringRepresentation();
		return this.stringRepresentation;
	}

	@Override
	public boolean equals(final Object other) {
		if (other == null) {
			return false;
		}
		if (other instanceof Clause == false) {
			return false;
		}
		return this.toString().contentEquals(other.toString());
	}

	/**
	 * Creates and returns a deep copy of this clause.<br>
	 * The {@link Clause#clashedWith} map (which is temporary information) will
	 * not be duplicated but will be left empty instead. Links to parent clauses
	 * and parent MGU will be copied as well.
	 * 
	 * @return A deep copy of this clause.
	 */
	public Clause duplicate() {
		return this.duplicate(null, this.leftParent, this.rightParent, this.parentMGU);
	}

	/**
	 * Creates and returns a deep copy of this clause.<br>
	 * The {@link Clause#clashedWith} map (which is temporary information) will
	 * not be duplicated but will be left empty instead.<br>
	 * <br>
	 * 
	 * NOTE: If you want to create a duplicate that originates from factoring,
	 * assign the left parent, and set rightParent and parentMGU to
	 * <code>null</code>.
	 * 
	 * @param excludedLiteral
	 *            The literal that should be excluded in the copy
	 * @param leftParent
	 *            The clause to use as left parent for the duplicate. May be
	 *            <code>null</code>.
	 * @param rightParent
	 *            The clause to use as right parent for the duplicate. May be
	 *            <code>null</code>.
	 * 
	 * @param parentMGU
	 *            The unifier to use for the parent MGU for the duplicate. May
	 *            be <code>null</code>.
	 * 
	 * @return A deep copy of this clause
	 */
	private Clause duplicate(final ClauseLiteral excludedLiteral, final Clause leftParent, final Clause rightParent,
			final Substitution parentMGU) {
		Clause copy = new Clause(leftParent, rightParent, parentMGU, SetOperations.union(this.literals,
				this.negatedLiterals));
		// copy.changed = true;
		// copy.recalculateHashCodeAndStringRepresentation();
		return copy;
	}

	/**
	 * Returns the left parent (one of the clauses that clashed and gave rise to
	 * this clause).
	 * 
	 * @return The left parent clause, or <code>null</code> if this clause was
	 *         in the initial clause set.
	 */
	public Clause getLeftParent() {
		return this.leftParent;
	}

	/**
	 * Returns the right parent (one of the clauses that clashed and gave rise
	 * to this clause).
	 * 
	 * @return The right parent clause, or <code>null</code> if this clause was
	 *         in the initial clause set.
	 */
	public Clause getRightParent() {
		return this.rightParent;
	}

	/**
	 * Returns the MGU which was used for deriving the clash between the two
	 * parent clauses.<br>
	 * The returned MGU may be empty.
	 * 
	 * @return The MGU which was used for deriving the clash between the two
	 *         parent clauses, or <code>null</code> if this clause was in the
	 *         original clause set.
	 */
	public Substitution getParentMGU() {
		return this.parentMGU;
	}

	/**
	 * Employs the resolution Factoring technique, creating the set of factors
	 * of this clause.<br>
	 * 
	 * All factors will list the same parent and same parent MGUs.
	 * 
	 * @return The set of clauses which are created by factoring this clause.
	 */
	public Set<Clause> factorize() {
		Set<Clause> factors = new HashSet<Clause>();
		this.factorizeRec(factors);
		return factors;
	}

	private void factorizeRec(final Set<Clause> allFactors) {
		Set<Clause> factors = new HashSet<Clause>();
		for (ClauseLiteral literal : this.literals) {
			Clause candidate = this.duplicate(literal, this, null, null);
			Clause unitClause = createUnitClause(literal);
			if (unitClause.subsumes(candidate)) {
				factors.add(candidate);
			}
		}
		if (factors.isEmpty() == false) {
			allFactors.addAll(factors);
			for (Clause factor : factors) {
				factor.factorizeRec(allFactors);
			}
		}
	}

}
