package at.cl.atp.foresolution.model.clause;

import java.util.Set;

import org.eclipse.emf.ecore.util.EcoreUtil;

import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.foresolution.transformations.substitution.Substitution;
import at.cl.atp.foresolution.transformations.substitution.mgu.Unification;
import at.cl.atp.foresolution.util.FormulaFormatter;
import at.cl.atp.foresolution.util.FormulaHelper;

public class ClauseLiteral {

	private final Formula formula;
	private Set<String> variables;
	private int variableOccurrenceCount;

	private int hashCode;
	// private boolean changed = true;
	private String stringRepresentation;

	public ClauseLiteral(final Formula formula) {
		this.formula = EcoreUtil.copy(formula);
		this.variables = FormulaHelper.getVariables(this.formula);
		this.variableOccurrenceCount = FormulaHelper.getVariableOccurrenceCount(this.formula);
		this.recalculateHashCodeAndStringRepresentation();
	}

	/**
	 * Checks if the root symbol of the formula stored in this literal is a
	 * negation.
	 * 
	 * @return <code>true</code> if the root symbol of this formula is a
	 *         negation, otherwise <code>false</code>.
	 */
	public boolean isNegated() {
		return this.formula instanceof NegatedFormula;
	}

	/**
	 * Returns the formula stored in this literal.<br>
	 * If the formula is negated, the negation symbol will be omitted in the
	 * returned formula.
	 * 
	 * @return The formula stored in the literal with leading negation omitted.
	 */
	private Formula getPlainFormula() {
		if (this.isNegated()) {
			return ((NegatedFormula) this.formula).getInnerFormula();
		} else {
			return this.formula;
		}
	}

	/**
	 * Checks if this literal unifies with the specified literal.
	 * 
	 * @param other
	 *            The literal to check if this literal unifies with it
	 * @return <code>true</code> if this literal unifies with the given literal,
	 *         otherwise <code>false</code>.
	 */
	public boolean unifiesWith(final ClauseLiteral other) {
		if (Unification.existsMGU(this.formula, other.formula)) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Checks if this literal clashes with the other given literal.
	 * 
	 * @param other
	 *            The other literal to clash with
	 * @return The most general unifier (MGU) that turns this literal into the
	 *         negation of the other literal.
	 */
	public Substitution getClashingMGU(final ClauseLiteral other) {
		Formula myFormula = null;
		Formula yourFormula = null;
		if (this.isNegated() && !other.isNegated()) {
			myFormula = this.getPlainFormula();
			yourFormula = other.formula;
		} else if (!this.isNegated() && other.isNegated()) {
			myFormula = this.formula;
			yourFormula = other.getPlainFormula();
		} else {
			return null;
		}
		return Unification.createMGU(myFormula, yourFormula);
	}

	private String toStringInternal() {
		return FormulaFormatter.format(this.formula);
	}

	/**
	 * Returns the formula that is owned by this literal.
	 * 
	 * @return The formula owned by this literal.
	 */
	public Formula getFormula() {
		return this.formula;
	}

	/**
	 * Returns the set of variables that occur in this formula.
	 * 
	 * @return The set of variables that occur in this formula.
	 */
	public Set<String> getVariables() {
		return this.variables;
	}

	/**
	 * Returns the number of variables occurring in this literal.<br>
	 * If the same variable appears multiple times, each of these occurrences
	 * will count as one occurrence nevertheless.
	 * 
	 * @return The total number of variables occurring in this formula.
	 */
	public int getVariableOccurrenceCount() {
		return this.variableOccurrenceCount;
	}

	/**
	 * Checks if this clause literal is syntactically equal to the given
	 * literal.
	 * 
	 * @param other
	 *            The other literal to compare this one to
	 * @return <code>true</code> if the other literal is syntactically
	 *         equivalent to this one, otherwise <code>false</code>.
	 */
	public boolean isSyntacticallyEqualTo(final ClauseLiteral other) {
		return this.toString().contentEquals(other.toString());
	}

	/**
	 * Creates and returns a deep copy of this literal.
	 * 
	 * @return A deep copy of this literal.
	 */
	public ClauseLiteral duplicate() {
		return new ClauseLiteral(this.formula);
	}

	/**
	 * Applies the given substitution to this literal.
	 * 
	 * @param sigma
	 *            The substitution to apply.
	 */
	public ClauseLiteral applySubstitution(final Substitution sigma) {
		if (sigma.isEmpty()) {
			return this.duplicate();
		}
		Formula formulaCopy = sigma.applyTo(EcoreUtil.copy(this.formula));
		return new ClauseLiteral(formulaCopy);
	}

	/**
	 * Recalculates the hashCode of this Clause, along with the result of the
	 * {@link Clause#toString()} operation.<br>
	 * Does nothing if the clause has not changed since the last time this
	 * operation was called.
	 */
	private void recalculateHashCodeAndStringRepresentation() {
		// if (!this.changed) {
		// return;
		// }
		this.stringRepresentation = this.toStringInternal();
		this.hashCode = this.stringRepresentation.hashCode();
		// this.changed = false;
	}

	@Override
	public int hashCode() {
		this.recalculateHashCodeAndStringRepresentation();
		return this.hashCode;
	}

	@Override
	public String toString() {
		this.recalculateHashCodeAndStringRepresentation();
		return this.stringRepresentation;
	}

	@Override
	public boolean equals(final Object other) {
		if (other == null) {
			return false;
		}
		if (other instanceof ClauseLiteral == false) {
			return false;
		}
		return this.toString().contentEquals(other.toString());
	}

}
