package at.cl.atp.foresolution.prover;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import at.cl.atp.foresolution.exceptions.ClashProducedBoxException;
import at.cl.atp.foresolution.model.clause.ClausalForm;
import at.cl.atp.foresolution.model.clause.ClausalFormBuilder;
import at.cl.atp.foresolution.model.clause.Clause;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.settings.Settings;

public class FirstOrderResolutionProver {

	public static ResolutionOutput run(final FormulaContainer formula) {
		ClausalForm clausalForm = ClausalFormBuilder.build(formula);
		return run(clausalForm);
	}

	public static ResolutionOutput run(final ClausalForm clausalForm) {
		tautologyElimination(clausalForm);
		if (clausalForm.isEmpty()) {
			return ResolutionOutput.createEmptySet();
		}
		if (Settings.useFactorization()) {
			Set<Clause> factors = new HashSet<Clause>();
			for (Clause clause : clausalForm.getClauses()) {
				factors.addAll(clause.factorize());
			}
			clausalForm.addAll(factors);
		}
		int clausesInLastIteration = 0;
		int derivedClausesCount = 0;
		int iterationIndex = 0;
		System.out.println();
		System.out.println("RESOLUTION START. Initial clause set:");
		for (Clause clause : clausalForm.getClauses()) {
			System.out.println(clause.toString());
		}
		while (clausalForm.size() > clausesInLastIteration) {
			System.out.println();
			System.out.println("-------------------------------------------------------------" + iterationIndex + " ("
					+ derivedClausesCount + ")");
			System.out.println();
			clausesInLastIteration = clausalForm.size();
			Set<Clause> derivedClauses = new HashSet<Clause>();
			for (Clause clauseA : clausalForm.getClauses()) {
				for (Clause clauseB : clausalForm.getClauses()) {
					if (clauseA != clauseB) {
						try {
							Set<Clause> resolvents = clauseA.clash(clauseB);
							derivedClauses.addAll(resolvents);
						} catch (ClashProducedBoxException e) {
							System.out.println(e.getMessage());
							System.out.println("RESOLUTION: Managed to derive BOX after deriving about "
									+ derivedClausesCount + " clauses in iteration # " + (iterationIndex + 1) + ".");
							return ResolutionOutput.createBox(e.getFirst(), e.getSecond(), e.getMgu());
						}
					}
				}
			}
			Set<Clause> newClauses = forwardSubsumption(clausalForm, derivedClauses);
			if (Settings.useFactorization()) {
				Set<Clause> factors = new HashSet<Clause>();
				for (Clause clause : clausalForm.getClauses()) {
					factors.addAll(clause.factorize());
				}
				newClauses.addAll(factors);
			}
			derivedClausesCount += newClauses.size();
			for (Clause clause : newClauses) {
				System.out.println(clause.toString());
			}
			clausalForm.addAll(newClauses);
			derivedClauses.clear();
			iterationIndex++;
		}
		System.out.println("RESOLUTION: Resolved all " + derivedClausesCount
				+ " derivable clauses without contradiction.");
		return ResolutionOutput.createNoContradiction();
	}

	private static void tautologyElimination(final ClausalForm clausalForm) {
		Iterator<Clause> iterator = clausalForm.clauseIterator();
		while (iterator.hasNext()) {
			Clause clause = iterator.next();
			if (clause.isEmpty() || clause.isTautology()) {
				iterator.remove();
			}
		}
	}

	private static Set<Clause> forwardSubsumption(final ClausalForm clausalForm, final Set<Clause> newClauses) {
		Set<Clause> resultSet = new HashSet<Clause>(newClauses);
		for (Clause existingClause : clausalForm.getClauses()) {
			for (Clause newClause : newClauses) {
				if (existingClause.subsumes(newClause)) {
					resultSet.remove(newClause);
				}
			}
		}
		return resultSet;
	}

}
