package at.cl.atp.foresolution.transformations;

import java.util.Collections;
import java.util.Set;

import at.cl.atp.fOResolution.Atom;
import at.cl.atp.fOResolution.BinaryOperation;
import at.cl.atp.fOResolution.FOResolutionFactory;
import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.Function;
import at.cl.atp.fOResolution.LowBindFormula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.fOResolution.Predicate;
import at.cl.atp.fOResolution.QuantifiedFormula;
import at.cl.atp.fOResolution.Quantifier;
import at.cl.atp.foresolution.model.container.FormulaContainer;

public class NNFTransformation implements FormulaTransformation {

	public static final NNFTransformation instance;

	static {
		instance = new NNFTransformation();
	}

	private NNFTransformation() {

	}

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		// we require that implications get eliminated before we apply NNF
		// transformation
		return Collections.singleton(TransformationType.IMPLICATION_ELIMINATION);
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.NEGATION_NORMAL_FORM)) {
			return;
		}
		TransformationProvider.resolvePreconditions(this, formula);
		formula.setFormula(this.doTransform(formula.getFormula()));
		formula.getAppliedTransformations().add(TransformationType.NEGATION_NORMAL_FORM);
	}

	private Formula doTransform(final Formula formula) {
		if (formula instanceof NegatedFormula) {
			return this.pushNegationInside(((NegatedFormula) formula).getInnerFormula());
		} else {
			if (formula instanceof LowBindFormula) {
				LowBindFormula f = (LowBindFormula) formula;
				f.setLeft(this.doTransform(f.getLeft()));
				f.setRight(this.doTransform(f.getRight()));
			} else if (formula instanceof QuantifiedFormula) {
				QuantifiedFormula f = (QuantifiedFormula) formula;
				f.setInnerFormula(this.doTransform(f.getInnerFormula()));
			} else if (formula instanceof Predicate) {
				// nothing to do for predicates
			} else if (formula instanceof Function) {
				// nothing to do for functions
			} else if (formula instanceof Atom) {
				// nothing to do for atoms
			}
			return formula;
		}
	}

	private Formula pushNegationInside(final Formula formula) {
		if (formula instanceof NegatedFormula) {
			// we have a double negation
			return this.doTransform(((NegatedFormula) formula).getInnerFormula());
		} else {
			if (formula instanceof LowBindFormula) {
				LowBindFormula f = (LowBindFormula) formula;
				// we need to change the operator accordingly
				if (f.getOperation() == BinaryOperation.CONJUNCTION) {
					f.setOperation(BinaryOperation.DISJUNCTION);
				} else if (f.getOperation() == BinaryOperation.DISJUNCTION) {
					f.setOperation(BinaryOperation.CONJUNCTION);
				}
				// we need to push the negation further inside
				f.setLeft(this.pushNegationInside(f.getLeft()));
				f.setRight(this.pushNegationInside(f.getRight()));
				return f;
			} else if (formula instanceof QuantifiedFormula) {
				QuantifiedFormula f = (QuantifiedFormula) formula;
				// we need to swap the quantifier symbol around
				if (f.getQuantifier() == Quantifier.EXISTENTIAL) {
					f.setQuantifier(Quantifier.UNIVERSAL);
				} else if (f.getQuantifier() == Quantifier.UNIVERSAL) {
					f.setQuantifier(Quantifier.EXISTENTIAL);
				}
				// we need to push the negation further inside
				f.setInnerFormula(this.pushNegationInside(f.getInnerFormula()));
				return f;
			} else if (formula instanceof Predicate) {
				// negate the predicate
				NegatedFormula negation = FOResolutionFactory.eINSTANCE.createNegatedFormula();
				negation.setInnerFormula(formula);
				return negation;
			} else if (formula instanceof Atom) {
				// negate the atom
				NegatedFormula negation = FOResolutionFactory.eINSTANCE.createNegatedFormula();
				negation.setInnerFormula(formula);
				return negation;
			}
			return formula;
		}
	}

}
