package at.cl.atp.foresolution.transformations;

import at.cl.atp.foresolution.model.container.FormulaContainer;

public class NormalFormTransformations {

	/**
	 * Transforms the given formula into negation normal form.
	 * 
	 * @param formula
	 *            The formula to transform. Will be changed.
	 * @return The original formula, transformed into NNF.
	 */
	public static void toNNF(final FormulaContainer formula) {
		applyTransformation(formula, TransformationType.NEGATION_NORMAL_FORM);
	}

	/**
	 * Performs standard implication elimination on the formula.
	 * 
	 * @param formula
	 *            The formula to transform. Will be changed.
	 * @return The original formula, with all implications eliminated.
	 */
	public static void implicationElimination(final FormulaContainer formula) {
		applyTransformation(formula, TransformationType.IMPLICATION_ELIMINATION);
	}

	/**
	 * Rectifies the formula.<br>
	 * That means that after applying this transformation, every variable will
	 * be bound to at most one quantifier (globally), regardless of its scope.
	 * 
	 * @param formula
	 *            The formula to rectify
	 */
	public static void rectify(final FormulaContainer formula) {
		applyTransformation(formula, TransformationType.RECTIFICATION);
	}

	/**
	 * Applies structural skolemization on the formula, eliminating all strong
	 * quantifiers.
	 * 
	 * @param formula
	 *            The formula to modify
	 */
	public static void structuralSkolemization(final FormulaContainer formula) {
		applyTransformation(formula, TransformationType.STRUCTURAL_SKOLEMIZATION);
	}

	/**
	 * Transforms the given formula into clausal form required for resolution.
	 * 
	 * @param formula
	 *            The formula to transform.
	 */
	public static void toClausalForm(final FormulaContainer formula) {
		applyTransformation(formula, TransformationType.CLAUSAL_FORM);
	}

	/**
	 * Applies the given transformation to the formula.
	 * 
	 * @param formula
	 *            The formula to transform. Will be modified.
	 * @param transformation
	 *            The transformation to apply.
	 * @return The formula, transformed by the given transformation type.
	 */
	public static void applyTransformation(final FormulaContainer formula, final TransformationType transformation) {
		TransformationProvider.getTransformation(transformation).transform(formula);
	}

}
