package at.cl.atp.foresolution.transformations;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import at.cl.atp.fOResolution.FOResolutionFactory;
import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.Function;
import at.cl.atp.fOResolution.LowBindFormula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.fOResolution.Predicate;
import at.cl.atp.fOResolution.QuantifiedFormula;
import at.cl.atp.fOResolution.Variable;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.settings.Settings;
import at.cl.atp.foresolution.transformations.substitution.Substitution;
import at.cl.atp.foresolution.util.FormulaHelper;

public class RectificationTransformation implements FormulaTransformation {

	public static final RectificationTransformation instance;

	static {
		instance = new RectificationTransformation();
	}

	private RectificationTransformation() {
	}

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		return Collections.emptySet();
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.RECTIFICATION)) {
			return;
		}
		if (Settings.noRenaming() == false) {
			this.renameFunctionsAndPredicates(formula, formula.getFormula());
			this.renameVariablesRec(formula, formula.getFormula());
		}
		formula.getAppliedTransformations().add(TransformationType.RECTIFICATION);
	}

	private void renameFunctionsAndPredicates(final FormulaContainer container, final Formula formula) {
		Set<String> functionNames = FormulaHelper.getFunctionSymbols(formula);
		Set<String> predicateNames = FormulaHelper.getPredicateSymbols(formula);
		Map<String, String> oldNameToNewName = new HashMap<String, String>();
		// create a new name and map the old name to it
		// note: we can use the same map for function and predicate names
		// here because the parser requires function names to start with
		// a lowercase letter while predicates must begin with an uppercase
		// letter, therefore we cannot have "collisions" between function
		// names and predicate names.
		for (String oldName : functionNames) {
			String newName = container.nextSkolemFunction();
			oldNameToNewName.put(oldName, newName);
		}
		for (String oldName : predicateNames) {
			String newName = container.nextSkolemPredicate();
			oldNameToNewName.put(oldName, newName);
		}
		// recursively walk through the tree, replacing function and predicate
		// names on the fly
		this.renameFunctionsAndPredicatesRec(formula, oldNameToNewName);
	}

	private void renameFunctionsAndPredicatesRec(final Formula formula, final Map<String, String> oldNameToNewName) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			this.renameFunctionsAndPredicatesRec(f.getInnerFormula(), oldNameToNewName);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			this.renameFunctionsAndPredicatesRec(f.getLeft(), oldNameToNewName);
			this.renameFunctionsAndPredicatesRec(f.getRight(), oldNameToNewName);
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			this.renameFunctionsAndPredicatesRec(f.getInnerFormula(), oldNameToNewName);
		} else if (formula instanceof Predicate) {
			Predicate f = (Predicate) formula;
			String newName = oldNameToNewName.get(f.getName());
			f.setName(newName);
			for (Formula argument : f.getArguments()) {
				this.renameFunctionsAndPredicatesRec(argument, oldNameToNewName);
			}
		} else if (formula instanceof Function) {
			Function f = (Function) formula;
			String newName = oldNameToNewName.get(f.getName());
			f.setName(newName);
			for (Formula argument : f.getArguments()) {
				this.renameFunctionsAndPredicatesRec(argument, oldNameToNewName);
			}
		}
	}

	private void renameVariablesRec(final FormulaContainer container, final Formula formula) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			// map each variable to a new identifier
			Map<String, String> oldToNewName = new HashMap<String, String>();
			for (String oldName : f.getVariables()) {
				oldToNewName.put(oldName, container.nextSkolemVariable());
			}
			// remove the old names from the quantifier
			f.getVariables().clear();
			// add the new names
			f.getVariables().addAll(oldToNewName.values());
			// perform a substitution of the variable names
			// in the scope of the quantifier
			Substitution substitution = new Substitution();
			for (Entry<String, String> renaming : oldToNewName.entrySet()) {
				// prepare a variable instance
				Variable variable = FOResolutionFactory.eINSTANCE.createVariable();
				variable.setName(renaming.getValue());
				// add the replacement of { oldName -> newName }
				substitution.addReplacement(renaming.getKey(), variable);
			}
			// perform the substitution on the inner formula of the quantifier
			Formula newInner = substitution.applyTo(f.getInnerFormula());
			f.setInnerFormula(newInner);
			this.renameVariablesRec(container, f.getInnerFormula());
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			// apply the transformation to both arguments
			this.renameVariablesRec(container, f.getLeft());
			this.renameVariablesRec(container, f.getRight());
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			// apply the transformation to the inner formula
			this.renameVariablesRec(container, f.getInnerFormula());
		} else if (formula instanceof Predicate) {
			Predicate f = (Predicate) formula;
			// apply the transformation to all arguments
			for (Formula argument : f.getArguments()) {
				this.renameVariablesRec(container, argument);
			}
		} else if (formula instanceof Function) {
			Function f = (Function) formula;
			// apply the transformation to all arguments
			for (Formula argument : f.getArguments()) {
				this.renameVariablesRec(container, argument);
			}
		}
	}

}
