package at.cl.atp.foresolution.transformations;

import java.util.HashSet;
import java.util.Set;

import at.cl.atp.fOResolution.FOResolutionFactory;
import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.Function;
import at.cl.atp.fOResolution.LowBindFormula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.fOResolution.QuantifiedFormula;
import at.cl.atp.fOResolution.Quantifier;
import at.cl.atp.fOResolution.Variable;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.transformations.substitution.Substitution;

public class StructuralSkolemization implements FormulaTransformation {

	public static final StructuralSkolemization instance;

	static {
		instance = new StructuralSkolemization();
	}

	private StructuralSkolemization() {
	}

	@Override
	public Set<TransformationType> getRequiredTransformations() {
		Set<TransformationType> preconditions = new HashSet<TransformationType>();
		preconditions.add(TransformationType.RECTIFICATION);
		preconditions.add(TransformationType.NEGATION_NORMAL_FORM);
		return preconditions;
	}

	@Override
	public void transform(final FormulaContainer formula) {
		if (formula.getAppliedTransformations().contains(TransformationType.STRUCTURAL_SKOLEMIZATION)) {
			return;
		}
		TransformationProvider.resolvePreconditions(this, formula);
		formula.setFormula(this.skolemizeRec(formula, formula.getFormula(), Quantifier.EXISTENTIAL,
				new HashSet<String>()));
		formula.getAppliedTransformations().add(TransformationType.STRUCTURAL_SKOLEMIZATION);
	}

	private Formula skolemizeRec(final FormulaContainer container, final Formula formula,
			final Quantifier strongQuantifier, final Set<String> boundVariables) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			if (f.getQuantifier() == strongQuantifier) {
				// prepare a substitution for the variables bound by the strong
				// quantifier
				// to new skolem functions that depend on the bound variables
				Substitution substitution = new Substitution();
				for (String variable : f.getVariables()) {
					Function skolemFunction = FOResolutionFactory.eINSTANCE.createFunction();
					skolemFunction.setName(container.nextSkolemFunction());
					// that function depends on all bound variables in our scope
					for (String boundVar : boundVariables) {
						Variable newVar = FOResolutionFactory.eINSTANCE.createVariable();
						newVar.setName(boundVar);
						skolemFunction.getArguments().add(newVar);
					}
					substitution.addReplacement(variable, skolemFunction);
				}
				// apply the substitution to the sub-formula of the quantifier
				Formula newInner = substitution.applyTo(f.getInnerFormula());
				// note that we return the substituted inner formula, NOT the
				// quantifier!
				// this way we remove the quantifier from the formula.
				return this.skolemizeRec(container, newInner, strongQuantifier, boundVariables);
			} else {
				// add all variables that are bound
				Set<String> newBoundVariables = new HashSet<String>(boundVariables);
				newBoundVariables.addAll(f.getVariables());
				f.setInnerFormula(this.skolemizeRec(container, f.getInnerFormula(), strongQuantifier, newBoundVariables));
				return f;
			}
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			f.setLeft(this.skolemizeRec(container, f.getLeft(), strongQuantifier, boundVariables));
			f.setRight(this.skolemizeRec(container, f.getRight(), strongQuantifier, boundVariables));
			return f;
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			// swap around the strong quantifier and apply to inner formula
			if (strongQuantifier == Quantifier.EXISTENTIAL) {
				f.setInnerFormula(this.skolemizeRec(container, f.getInnerFormula(), Quantifier.UNIVERSAL,
						boundVariables));
			} else {
				f.setInnerFormula(this.skolemizeRec(container, f.getInnerFormula(), Quantifier.EXISTENTIAL,
						boundVariables));
			}
		}
		// in all other cases (predicates, functions, atoms) we leave
		// the formula unchanged
		return formula;
	}

}
