package at.cl.atp.foresolution.transformations.substitution.mgu;

import java.util.List;

import org.eclipse.emf.ecore.util.EcoreUtil;

import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.Function;
import at.cl.atp.fOResolution.Predicate;
import at.cl.atp.fOResolution.Variable;
import at.cl.atp.foresolution.exceptions.EqualityToSubstitutionException;
import at.cl.atp.foresolution.transformations.substitution.Substitution;
import at.cl.atp.foresolution.util.FormulaFormatter;
import at.cl.atp.foresolution.util.FormulaHelper;

public class Equality {

	private Formula left;
	private Formula right;

	public Equality(final Formula left, final Formula right) {
		this.left = left;
		this.right = right;
	}

	public Formula getLeft() {
		return this.left;
	}

	public Variable getLeftAsVariable() {
		return (Variable) this.left;
	}

	public Formula getRight() {
		return this.right;
	}

	public boolean hasVariableOnEitherSide() {
		return this.hasVariableOnLeftHandSide() || this.hasVariableOnRightHandSide();
	}

	public boolean hasVariableOnLeftHandSide() {
		return this.left instanceof Variable;
	}

	public boolean hasVariableOnRightHandSide() {
		return this.right instanceof Variable;
	}

	public boolean isSyntacticEquality() {
		return EcoreUtil.equals(this.left, this.right);
		// String leftString = FormulaFormatter.format(this.left);
		// String rightString = FormulaFormatter.format(this.right);
		// return leftString.contentEquals(rightString);
	}

	public boolean hasLeadingFunction() {
		if (this.left instanceof Function && this.right instanceof Function) {
			return true;
		} else if (this.left instanceof Predicate && this.right instanceof Predicate) {
			return true;
		}
		return false;
	}

	public boolean isLeadingFunctionSymbolEqual() {
		if (this.left instanceof Function && this.right instanceof Function) {
			String leftName = ((Function) this.left).getName();
			String rightName = ((Function) this.right).getName();
			return leftName.contentEquals(rightName);
		} else if (this.left instanceof Predicate && this.right instanceof Predicate) {
			String leftName = ((Predicate) this.left).getName();
			String rightName = ((Predicate) this.right).getName();
			return leftName.contentEquals(rightName);
		} else {
			return false;
		}
	}

	public List<Formula> getArgumentsOfLeftFunction() {
		if (this.left instanceof Function) {
			Function f = (Function) this.left;
			return f.getArguments();
		} else if (this.left instanceof Predicate) {
			Predicate p = (Predicate) this.left;
			return p.getArguments();
		}
		return null;
	}

	public List<Formula> getArgumentsOfRightFunction() {
		if (this.right instanceof Function) {
			Function f = (Function) this.right;
			return f.getArguments();
		} else if (this.right instanceof Predicate) {
			Predicate p = (Predicate) this.right;
			return p.getArguments();
		}
		return null;
	}

	public void swapLeftAndRight() {
		Formula temp = this.left;
		this.left = this.right;
		this.right = temp;
	}

	public boolean occursCheck() {
		if (this.hasVariableOnLeftHandSide() == false) {
			return false;
		}
		String variableName = this.getLeftAsVariable().getName();
		return FormulaHelper.occursCheck(variableName, this.right);
	}

	/**
	 * Applies the given substitution to both sides of the equality.
	 * 
	 * @param sub
	 *            the substitution to apply
	 */
	public void applySubstitution(final Substitution sub) {
		this.left = sub.applyTo(this.left);
		this.right = sub.applyTo(this.right);
	}

	/**
	 * Transforms the equality into a substitution.<br>
	 * This is only possible if the left-hand side is a variable.
	 * 
	 * @return The substitution which substitutes the left-hand variable of this
	 *         equality with the formula on the right-hand side of this
	 *         equality.
	 * @throws EqualityToSubstitutionException
	 *             : Thrown if this equality does not have a single variable on
	 *             its left-hand side.
	 */
	public Substitution toSubstitution() throws EqualityToSubstitutionException {
		if (!this.hasVariableOnLeftHandSide()) {
			throw new EqualityToSubstitutionException(this);
		}
		Substitution substitution = new Substitution();
		substitution.addReplacement(this.getLeftAsVariable().getName(), this.right);
		return substitution;
	}

	@Override
	public String toString() {
		return "EQ[" + FormulaFormatter.format(this.left) + " = " + FormulaFormatter.format(this.right) + "]";
	}

}
