package at.cl.atp.foresolution.transformations.substitution.mgu;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import at.cl.atp.fOResolution.Formula;
import at.cl.atp.foresolution.exceptions.NotUnifiableException;
import at.cl.atp.foresolution.transformations.substitution.Substitution;

public class UnificationProblem {

	private final boolean leftToRigthRestriction;

	private final Set<Equality> equalities;
	private final Set<Equality> solvedEqualities;

	private final Formula originalLeft;
	private final Formula originalRight;

	public UnificationProblem(final Formula left, final Formula right, final boolean leftToRightRestriction) {
		this.equalities = new HashSet<Equality>();
		this.solvedEqualities = new HashSet<Equality>();
		this.leftToRigthRestriction = leftToRightRestriction;
		this.originalLeft = left;
		this.originalRight = right;
		this.equalities.add(new Equality(left, right));
	}

	public Substitution solve() throws NotUnifiableException {
		while (!this.equalities.isEmpty()) {
			Set<Equality> set = new HashSet<Equality>(this.equalities);
			for (Equality equality : set) {
				if (equality.isSyntacticEquality()) {
					// case { t = t } -> delete
					this.equalities.remove(equality);
				} else if (equality.hasLeadingFunction()) {
					if (equality.isLeadingFunctionSymbolEqual()) {
						// case { f(s0,s1,...,sk) = f(t0,t1,...,tk }
						// -> unify arguments pairwise via index
						List<Formula> leftArguments = equality.getArgumentsOfLeftFunction();
						List<Formula> rightArguments = equality.getArgumentsOfRightFunction();
						if (leftArguments.size() != rightArguments.size()) {
							// case { f(s0,s1,...,sk) = f(t0,t1,...tm) }
							// and k != m --> not unifiable
							throw new NotUnifiableException(this.originalLeft, this.originalRight);
						}
						int size = leftArguments.size();
						for (int index = 0; index < size; index++) {
							Formula leftArg = leftArguments.get(index);
							Formula rightArg = rightArguments.get(index);
							Equality newEquality = new Equality(leftArg, rightArg);
							this.equalities.add(newEquality);
						}
						this.equalities.remove(equality);
					} else {
						// case { f(s0,s1,...,sk) = g(t0,t1,...,tk }
						// -> not unifiable
						throw new NotUnifiableException(this.originalLeft, this.originalRight);
					}
				} else if (equality.hasVariableOnEitherSide()) {
					if (equality.hasVariableOnRightHandSide()) {
						// case { f(s0,s1,...,sk) = x }
						if (this.leftToRigthRestriction && !equality.hasVariableOnLeftHandSide()
								&& equality.hasVariableOnRightHandSide()) {
							// with the "left to right restriction",
							// we may not swap left and right hand
							// side of the equality, therefore we
							// cannot unify here
							throw new NotUnifiableException(this.originalLeft, this.originalRight);
						}
						// -> swap left and right
						equality.swapLeftAndRight();
					}
					if (equality.occursCheck()) {
						// case { {x=f(s0,s1,...,sk) } and x in vars(f)
						throw new NotUnifiableException(this.originalLeft, this.originalRight);
					} else {
						// important note: we must not apply the substitution
						// given by the equality to the equality itself,
						// otherwise we end up with an empty MGU!
						this.equalities.remove(equality);
						this.applySubstitution(equality.toSubstitution());
						this.solvedEqualities.add(equality);
					}
				} else {
					throw new NotUnifiableException(this.originalLeft, this.originalRight);
				}
			}
		}
		return this.constructMGU();
	}

	/**
	 * Constructs a substitution consisting of all solved equalities.
	 * 
	 * @return The Most General Unifier (MGU) substitution
	 */
	private Substitution constructMGU() {
		Substitution mgu = new Substitution();
		for (Equality equality : this.solvedEqualities) {
			mgu.union(equality.toSubstitution());
		}
		return mgu;
	}

	/**
	 * Applies the given substitution to all equalities (solved and non-solved)
	 * in this unification problem.
	 * 
	 * @param sub
	 *            The substitution to apply
	 */
	private void applySubstitution(final Substitution sub) {
		for (Equality equality : this.equalities) {
			equality.applySubstitution(sub);
		}
		for (Equality equality : this.solvedEqualities) {
			equality.applySubstitution(sub);
		}
	}
}
