package at.cl.atp.foresolution.util;

import java.util.HashSet;
import java.util.Set;

import at.cl.atp.fOResolution.Constant;
import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.Function;
import at.cl.atp.fOResolution.LowBindFormula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.fOResolution.Predicate;
import at.cl.atp.fOResolution.QuantifiedFormula;
import at.cl.atp.fOResolution.Variable;
import at.cl.atp.foresolution.model.clause.Clause;
import at.cl.atp.foresolution.model.clause.ClauseLiteral;
import at.cl.atp.foresolution.transformations.substitution.Substitution;
import at.cl.atp.foresolution.transformations.substitution.mgu.Unification;

public class FormulaHelper {

	/**
	 * Returns all free variables that occur in the formula.<br>
	 * A variable is considered to be <i>free</i> if there is no quantifier in
	 * the formula that binds the variable to it.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of all bound variables within the formula.
	 */
	public static Set<String> getFreeVariables(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		Set<String> boundVariables = new HashSet<String>();
		getFreeVariablesRec(formula, resultSet, boundVariables);
		return resultSet;
	}

	private static void getFreeVariablesRec(final Formula formula, final Set<String> freeVariables,
			final Set<String> boundInScope) {
		if (formula instanceof Variable) {
			Variable variable = (Variable) formula;
			if (!boundInScope.contains(variable.getName())) {
				freeVariables.add(variable.getName());
			}
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			Set<String> newScope = new HashSet<String>(boundInScope);
			newScope.addAll(f.getVariables());
			getFreeVariablesRec(f.getInnerFormula(), freeVariables, newScope);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			getFreeVariablesRec(f.getLeft(), freeVariables, boundInScope);
			getFreeVariablesRec(f.getRight(), freeVariables, boundInScope);
		} else if (formula instanceof NegatedFormula) {
			getFreeVariablesRec(((NegatedFormula) formula).getInnerFormula(), freeVariables, boundInScope);
		} else if (formula instanceof Predicate) {
			for (Formula argument : ((Predicate) formula).getArguments()) {
				getFreeVariablesRec(argument, freeVariables, boundInScope);
			}
		} else if (formula instanceof Function) {
			for (Formula argument : ((Function) formula).getArguments()) {
				getFreeVariablesRec(argument, freeVariables, boundInScope);
			}
		}
	}

	/**
	 * Returns all bound variables within the formula.<br>
	 * A variable is considered to be <i>bound</i> if there exists a quantifier
	 * in the formula that binds that variable to itself.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of bound variables within the formula
	 */
	public static Set<String> getBoundVariables(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getBoundVariablesRec(formula, resultSet);
		return resultSet;
	}

	private static void getBoundVariablesRec(final Formula formula, final Set<String> resultSet) {
		// note: we are only interested in the quantors here.
		// since they cannot appear inside functions or
		// predicates, we can safely ignore those constructs
		// if we encounter them. Since variables are bound
		// by quantors only (not by themselves) we can ignore
		// them as well.
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			resultSet.addAll(f.getVariables());
			getBoundVariablesRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			getBoundVariablesRec(f.getLeft(), resultSet);
			getBoundVariablesRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			getBoundVariablesRec(((NegatedFormula) formula).getInnerFormula(), resultSet);
		}
	}

	/**
	 * Gets all variables used in the formula.<br>
	 * Formally, this returns the <i>union</i> of all free variables and all
	 * bound variables in the formula.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of all variables occurring in the formula.
	 */
	public static Set<String> getVariables(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getVariablesRec(formula, resultSet);
		return resultSet;
	}

	private static void getVariablesRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof Variable) {
			Variable variable = (Variable) formula;
			resultSet.add(variable.getName());
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getVariablesRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			getVariablesRec(f.getLeft(), resultSet);
			getVariablesRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			getVariablesRec(((NegatedFormula) formula).getInnerFormula(), resultSet);
		} else if (formula instanceof Predicate) {
			for (Formula argument : ((Predicate) formula).getArguments()) {
				getVariablesRec(argument, resultSet);
			}
		} else if (formula instanceof Function) {
			for (Formula argument : ((Function) formula).getArguments()) {
				getVariablesRec(argument, resultSet);
			}
		}
	}

	/**
	 * Returns the set of function names contained in the formula.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of function names occurring in the formula.
	 */
	public static Set<String> getFunctionSymbols(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getFunctionSymbolsRec(formula, resultSet);
		return resultSet;
	}

	private static void getFunctionSymbolsRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getFunctionSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			getFunctionSymbolsRec(f.getLeft(), resultSet);
			getFunctionSymbolsRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			getFunctionSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof Predicate) {
			Predicate f = (Predicate) formula;
			for (Formula argument : f.getArguments()) {
				getFunctionSymbolsRec(argument, resultSet);
			}
		} else if (formula instanceof Function) {
			Function f = (Function) formula;
			resultSet.add(f.getName());
			for (Formula argument : f.getArguments()) {
				getFunctionSymbolsRec(argument, resultSet);
			}
		}
	}

	/**
	 * Returns the set of predicate names contained in the formula.
	 * 
	 * @param formula
	 *            The formula to scan
	 * @return The set of predicate names occurring in the formula.
	 */
	public static Set<String> getPredicateSymbols(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getPredicateSymbolsRec(formula, resultSet);
		return resultSet;
	}

	private static void getPredicateSymbolsRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getPredicateSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			getPredicateSymbolsRec(f.getLeft(), resultSet);
			getPredicateSymbolsRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula f = (NegatedFormula) formula;
			getPredicateSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof Predicate) {
			Predicate f = (Predicate) formula;
			resultSet.add(f.getName());
			for (Formula argument : f.getArguments()) {
				getPredicateSymbolsRec(argument, resultSet);
			}
		}
	}

	/**
	 * Checks if the given variable occurs in the formula or not.
	 * 
	 * @param variable
	 *            The variable to search for
	 * @param formula
	 *            The formula to check
	 * @return <code>true</code> if the variable occurs in the formula,
	 *         otherwise <code>false</code>.
	 */
	public static boolean occursCheck(final String variable, final Formula formula) {
		Set<String> variables = getVariables(formula);
		return variables.contains(variable);
	}

	/**
	 * Checks if the given formula is a sentence (i.e. has no free variables).
	 * 
	 * @param formula
	 *            The formula to check
	 * @return <code>true</code> if formula is sentence, otherwise
	 *         <code>false</code>.
	 */
	public static boolean isSentence(final Formula formula) {
		return getFreeVariables(formula).isEmpty();
	}

	/**
	 * Returns the set of constant symbols in the formula.
	 * 
	 * @param formula
	 *            The formula to analyze
	 * @return The set of constants appearing within the formula.
	 */
	public static Set<String> getConstantSymbols(final Formula formula) {
		Set<String> resultSet = new HashSet<String>();
		getConstantSymbolsRec(formula, resultSet);
		return resultSet;
	}

	private static void getConstantSymbolsRec(final Formula formula, final Set<String> resultSet) {
		if (formula instanceof Constant) {
			Constant constant = (Constant) formula;
			resultSet.add(constant.getName());
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			getConstantSymbolsRec(f.getInnerFormula(), resultSet);
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			getConstantSymbolsRec(f.getLeft(), resultSet);
			getConstantSymbolsRec(f.getRight(), resultSet);
		} else if (formula instanceof NegatedFormula) {
			getConstantSymbolsRec(((NegatedFormula) formula).getInnerFormula(), resultSet);
		} else if (formula instanceof Predicate) {
			for (Formula argument : ((Predicate) formula).getArguments()) {
				getConstantSymbolsRec(argument, resultSet);
			}
		} else if (formula instanceof Function) {
			for (Formula argument : ((Function) formula).getArguments()) {
				getConstantSymbolsRec(argument, resultSet);
			}
		}
	}

	/**
	 * Counts all variable occurrences within the formula.<br>
	 * If the same variable appears multiple times within the formula, each
	 * occurrence will be counted. Variable bindings that occur inside a
	 * quantifier will not be counted, only variable references are of interest.
	 * 
	 * @param formula
	 *            The formula to analyze
	 * @return The total number of variable occurrences within the formula.
	 */
	public static int getVariableOccurrenceCount(final Formula formula) {
		if (formula instanceof Variable) {
			return 1;
		} else if (formula instanceof QuantifiedFormula) {
			QuantifiedFormula f = (QuantifiedFormula) formula;
			return getVariableOccurrenceCount(f.getInnerFormula());
		} else if (formula instanceof LowBindFormula) {
			LowBindFormula f = (LowBindFormula) formula;
			int temp = getVariableOccurrenceCount(f.getLeft());
			temp += getVariableOccurrenceCount(f.getRight());
			return temp;
		} else if (formula instanceof NegatedFormula) {
			return getVariableOccurrenceCount(((NegatedFormula) formula).getInnerFormula());
		} else if (formula instanceof Predicate) {
			int temp = 0;
			for (Formula argument : ((Predicate) formula).getArguments()) {
				temp += getVariableOccurrenceCount(argument);
			}
			return temp;
		} else if (formula instanceof Function) {
			int temp = 0;
			for (Formula argument : ((Function) formula).getArguments()) {
				temp += getVariableOccurrenceCount(argument);
			}
			return temp;
		}
		return 0;
	}

	/**
	 * Forms a debug message which explains a clash of the given clauses and
	 * their literals.
	 * 
	 * @param first
	 *            The first clause which participated in a successful clash
	 * @param second
	 *            The second clause which participated in a successful clash
	 * @param firstLiteral
	 *            The literal of the first clause which clashed
	 * @param secondLiteral
	 *            The literal of the second clause which clashed
	 * @return The explanation debug string for the clash.
	 */
	public static String clashDebugMessage(final Clause first, final Clause second, final ClauseLiteral firstLiteral,
			final ClauseLiteral secondLiteral) {
		Formula firstFormula = null;
		Formula secondFormula = null;
		if (firstLiteral.isNegated()) {
			firstFormula = ((NegatedFormula) firstLiteral.getFormula()).getInnerFormula();
		} else {
			firstFormula = firstLiteral.getFormula();
		}
		if (secondLiteral.isNegated()) {
			secondFormula = ((NegatedFormula) secondLiteral.getFormula()).getInnerFormula();
		} else {
			secondFormula = secondLiteral.getFormula();
		}
		System.out.println("Building verification MGU...");
		System.out.println("First clause: " + first.toString());
		System.out.println("Second clause: " + second.toString());
		System.out.println("First literal: " + firstLiteral.toString());
		System.out.println("Second literal: " + secondLiteral.toString());
		Substitution mgu = Unification.createMGU(firstFormula, secondFormula);
		if (mgu == null) {
			System.out.println("ERROR ERROR ERROR");
			System.out.println("Could not find MGU for supposingly clashing literals:");
			System.out.println("First: " + firstLiteral.toString());
			System.out.println("Second: " + secondLiteral.toString());
		}
		StringBuilder builder = new StringBuilder();
		builder.append("CLASH :: ");
		builder.append(first.toString());
		builder.append(" with ");
		builder.append(second.toString());
		builder.append(" on ");
		builder.append(firstLiteral.toString());
		builder.append(" and ");
		builder.append(secondLiteral.toString());
		builder.append(" with MGU ");
		builder.append(mgu.toString());
		return builder.toString();
	}
}
