package resolution;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.util.EcoreUtil;
import org.junit.Test;

import util.TestHelper;
import at.cl.atp.fOResolution.FOResolutionFactory;
import at.cl.atp.fOResolution.Formula;
import at.cl.atp.fOResolution.NegatedFormula;
import at.cl.atp.foresolution.exceptions.ClashProducedBoxException;
import at.cl.atp.foresolution.latexExporter.LatexExporter;
import at.cl.atp.foresolution.model.clause.ClausalForm;
import at.cl.atp.foresolution.model.clause.ClausalFormBuilder;
import at.cl.atp.foresolution.model.clause.Clause;
import at.cl.atp.foresolution.model.container.FormulaContainer;
import at.cl.atp.foresolution.prover.ResolutionOutput;
import at.cl.atp.foresolution.prover.ResolutionResult;
import at.cl.atp.foresolution.settings.Settings;
import at.cl.atp.foresolution.transformations.NormalFormTransformations;
import at.cl.atp.foresolution.util.FormulaFormatter;
import at.cl.atp.foresolution.util.FormulaHelper;

public class ImmutabilityTest {

	private final Map<Clause, String> clauseToStringMap = new HashMap<Clause, String>();
	private final Set<Clause> originalClauseSet = new HashSet<Clause>();

	@Test
	public void immutabilityTest() {
		long timeBefore = System.currentTimeMillis();
		Formula formula = TestHelper.load("killerExample");
		long timeAfterParsing = System.currentTimeMillis();
		System.out.println();
		System.out.println();
		System.out.println();
		System.out.println("___RUNNING_TEST_IMMUTABILITY_TEST__________________");
		FormulaContainer container = new FormulaContainer(formula);
		System.out.println("ORIGINAL FORMULA");
		System.out.println("\t" + FormulaFormatter.format(formula));
		System.out.println("\tFree Variables:  " + FormulaHelper.getFreeVariables(formula).toString());
		System.out.println("\tBound Variables: " + FormulaHelper.getBoundVariables(formula).toString());
		System.out.println("\tConstant names:  " + FormulaHelper.getConstantSymbols(formula).toString());
		System.out.println("\tFunction names:  " + FormulaHelper.getFunctionSymbols(formula).toString());
		System.out.println("\tPredicate names: " + FormulaHelper.getPredicateSymbols(formula).toString());
		System.out.println();
		Formula f = null;
		if (Settings.negateInput()) {
			NegatedFormula negation = FOResolutionFactory.eINSTANCE.createNegatedFormula();
			negation.setInnerFormula(EcoreUtil.copy(formula));
			f = negation;
		} else {
			f = formula;
		}
		container.setFormula(f);
		NormalFormTransformations.toNNF(container);
		System.out.println("Negated and transformed to NNF:");
		System.out.println("\t" + container.toString());
		System.out.println("Rectified Refutational-Skolemized CNF:");
		NormalFormTransformations.toClausalForm(container);
		System.out.println("\t" + container.toString());
		ClausalForm clausalForm = ClausalFormBuilder.build(container);
		this.originalClauseSet.addAll(clausalForm.getClauses());
		System.out.println("___ORIGINAL_CLAUSE_SET_BEFORE_RESOLUTION_EXECUTION_______________");
		for (Clause clause : this.originalClauseSet) {
			System.out.println(clause.toString());
		}
		System.out.println("Clausal Form:");
		System.out.println("\t" + clausalForm.toString());
		ResolutionOutput outcome = this.runCheckedResolution(clausalForm);
		ResolutionResult result = outcome.getResult();
		long timeAfter = System.currentTimeMillis();
		System.out.println();
		System.out.println("___RESULT________________________________________");
		System.out.println("Result of resolution: " + result);
		System.out.println();
		System.out.println("___TIME__________________________________________");
		System.out.println("Parse time: " + (timeAfterParsing - timeBefore) + "ms.");
		System.out.println("Resolution: " + (timeAfter - timeAfterParsing) + "ms.");
		System.out.println("Total time: " + (timeAfter - timeBefore) + "ms.");
		long memTotal = Runtime.getRuntime().totalMemory();
		long memMax = Runtime.getRuntime().maxMemory();
		long memUsed = memTotal - Runtime.getRuntime().freeMemory();
		System.out.println();
		System.out.println("___MEMORY________________________________________");
		System.out.println("Available: " + TestHelper.toMegabytes(memMax));
		System.out.println("Allocated: " + TestHelper.toMegabytes(memTotal));
		System.out.println("Used:      " + TestHelper.toMegabytes(memUsed));
		assertEquals(ResolutionResult.BOX, result);
		for (Clause axiom : this.originalClauseSet) {
			System.out.println(axiom.toString());
		}
		this.checkOrigins(outcome.getLeftParentOfBox());
		this.checkOrigins(outcome.getRightParentOfBox());
		System.out.println("___ORIGINAL_CLAUSE_SET_AFTER_RESOLUTION_EXECUTION_______________");
		for (Clause clause : this.originalClauseSet) {
			System.out.println(clause.toString());
		}
		LatexExporter.exportToLatex(outcome, formula);
	}

	private void checkOrigins(final Clause clause) {
		if (clause.getLeftParent() != null) {
			this.checkOrigins(clause.getLeftParent());
			if (clause.getRightParent() != null) {
				this.checkOrigins(clause.getRightParent());
			}
		} else {
			// we're at a root clause which MUST have been
			// in the original clause set!
			assertTrue("ROOT CLAUSE " + clause.toString() + " NOT CONTAINED IN ORIGINAL CLAUSE SET!",
					this.originalClauseSet.contains(clause));
		}

	}

	private ResolutionOutput runCheckedResolution(final ClausalForm clausalForm) {
		this.tautologyElimination(clausalForm);
		if (clausalForm.isEmpty()) {
			return ResolutionOutput.createEmptySet();
		}
		if (Settings.useFactorization()) {
			Set<Clause> factors = new HashSet<Clause>();
			for (Clause clause : clausalForm.getClauses()) {
				factors.addAll(clause.factorize());
			}
			clausalForm.addAll(factors);
		}
		int clausesInLastIteration = 0;
		int derivedClausesCount = 0;
		int iterationIndex = 0;
		System.out.println();
		System.out.println("RESOLUTION START. Initial clause set:");
		for (Clause clause : clausalForm.getClauses()) {
			System.out.println(clause.toString());
			this.mapAndCheck(clause);
		}
		while (clausalForm.size() > clausesInLastIteration) {
			System.out.println();
			System.out.println("-------------------------------------------------------------" + iterationIndex + " ("
					+ derivedClausesCount + ")");
			System.out.println();
			clausesInLastIteration = clausalForm.size();
			Set<Clause> derivedClauses = new HashSet<Clause>();
			for (Clause clauseA : clausalForm.getClauses()) {
				for (Clause clauseB : clausalForm.getClauses()) {
					if (clauseA != clauseB) {
						try {
							Set<Clause> resolvents = clauseA.clash(clauseB);
							derivedClauses.addAll(resolvents);
						} catch (ClashProducedBoxException e) {
							System.out.println(e.getMessage());
							System.out.println("RESOLUTION: Managed to derive BOX after deriving about "
									+ derivedClausesCount + " clauses in iteration # " + (iterationIndex + 1) + ".");
							return ResolutionOutput.createBox(e.getFirst(), e.getSecond(), e.getMgu());
						}
					}
				}
			}
			Set<Clause> newClauses = this.forwardSubsumption(clausalForm, derivedClauses);
			if (Settings.useFactorization()) {
				Set<Clause> factors = new HashSet<Clause>();
				for (Clause clause : clausalForm.getClauses()) {
					factors.addAll(clause.factorize());
				}
				newClauses.addAll(factors);
			}
			derivedClausesCount += newClauses.size();
			for (Clause clause : newClauses) {
				this.checkOrigins(clause);
			}
			for (Clause clause : newClauses) {
				System.out.println("Checking origins of: " + clause.toString());
				System.out.println("\tLeft Parent: " + clause.getLeftParent());
				System.out.println("\tRight Parent: " + clause.getRightParent());
				this.checkOrigins(clause);
			}
			clausalForm.addAll(newClauses);
			derivedClauses.clear();
			for (Clause clause : clausalForm.getClauses()) {
				this.mapAndCheck(clause);
			}
			System.out.println();
			iterationIndex++;
		}
		System.out.println("RESOLUTION: Resolved all " + derivedClausesCount
				+ " derivable clauses without contradiction.");
		return ResolutionOutput.createNoContradiction();
	}

	private Set<Clause> forwardSubsumption(final ClausalForm clausalForm, final Set<Clause> newClauses) {
		Set<Clause> resultSet = new HashSet<Clause>(newClauses);
		for (Clause existingClause : clausalForm.getClauses()) {
			for (Clause newClause : newClauses) {
				if (existingClause.subsumes(newClause)) {
					resultSet.remove(newClause);
				}
			}
		}
		return resultSet;
	}

	private void tautologyElimination(final ClausalForm clausalForm) {
		Iterator<Clause> iterator = clausalForm.clauseIterator();
		while (iterator.hasNext()) {
			Clause clause = iterator.next();
			if (clause.isEmpty() || clause.isTautology()) {
				iterator.remove();
			}
		}
	}

	private void mapAndCheck(final Clause clause) {
		if (this.clauseToStringMap.containsKey(clause)) {
			assertTrue("CLAUSE " + clause.toString() + " HAS CHANGED! WAS: " + this.clauseToStringMap.get(clause),
					this.clauseToStringMap.get(clause).contentEquals(clause.toString()));
			System.out.print(".");
		} else {
			this.clauseToStringMap.put(clause, clause.toString());
		}
	}
}
