package at.cl.atp.foresolution.latexExporter

import at.cl.atp.foresolution.prover.ResolutionOutput
import java.io.File
import java.io.PrintWriter
import at.cl.atp.fOResolution.Formula
import at.cl.atp.foresolution.model.clause.Clause
import java.util.Map
import at.cl.atp.foresolution.transformations.substitution.Substitution
import java.util.HashMap
import at.cl.atp.foresolution.model.container.FormulaContainer
import org.eclipse.emf.ecore.util.EcoreUtil
import at.cl.atp.foresolution.transformations.NormalFormTransformations
import at.cl.atp.foresolution.model.clause.ClausalFormBuilder
import at.cl.atp.foresolution.model.clause.ClausalForm
import at.cl.atp.fOResolution.FOResolutionFactory
import at.cl.atp.fOResolution.NegatedFormula

class LatexExporter {
	
	static val Map<Integer, Substitution> indexSigmaMap = new HashMap<Integer, Substitution>();
	static val Map<Substitution, Integer> sigmaIndexMap = new HashMap<Substitution, Integer>();
	static var mguIndex = 0;
	
	
	def static void exportToLatex(ResolutionOutput output, Formula formula){
		sigmaIndexMap.clear;
		indexSigmaMap.clear;
		mguIndex = 1;
		// iterate over the output clauses and index the substitutions
		sigmaIndexMap.put(output.mgu, 0);
		indexSigmaMap.put(0, output.mgu);
		initializeSigmaIndexMap(output.leftParentOfBox);
		initializeSigmaIndexMap(output.rightParentOfBox);		
		val File texFile = new File("output.tex");
		if(texFile.exists == false){
			texFile.createNewFile;
		}
		val PrintWriter writer = new PrintWriter(texFile);
		var String contents = generateContents(output, formula);
		writer.write(contents);
		writer.flush;
		writer.close;
	}
	
	def private static void initializeSigmaIndexMap(Clause clause){
		if(clause.parentMGU != null){
			if(sigmaIndexMap.containsKey(clause.parentMGU) == false){
				indexSigmaMap.put(mguIndex, clause.parentMGU);
				sigmaIndexMap.put(clause.parentMGU, mguIndex);
				mguIndex = mguIndex + 1;
			}
		}
		if(clause.leftParent != null){
			initializeSigmaIndexMap(clause.leftParent);
		}
		if(clause.rightParent != null){
			initializeSigmaIndexMap(clause.rightParent);
		}
	}
	
	def private static String generateContents(ResolutionOutput output, Formula formula){
		val NegatedFormula negation = FOResolutionFactory::eINSTANCE.createNegatedFormula;
		negation.innerFormula = EcoreUtil::copy(formula);
		val FormulaContainer container = new FormulaContainer(negation);
		NormalFormTransformations::rectify(container);
		NormalFormTransformations::toNNF(container);
		NormalFormTransformations::structuralSkolemization(container);
		NormalFormTransformations::toClausalForm(container);
		val ClausalForm clausalForm = ClausalFormBuilder::build(container);
		return '''
			\documentclass[border=1cm,varwidth=\maxdimen]{standalone}

			\usepackage{bussproofs}
			\usepackage{amssymb}
			
			
			\begin{document}
			
			\textbf{The formula}\\
			This is the proof of the following formula:\\
			\begin{center}
			«LatexFormatter::format(formula)»
			\end{center}
			\vspace{1cm}
			\textbf{The NNF and Clausal Forms}\\
			The skolemized Negation Normal Form of the negated input:
			\begin{center}
			«LatexFormatter::format(container.formula)»
			\end{center}
			
			\noindent The corresponding clausal form:
			\begin{center}
			«LatexFormatter::format(clausalForm)»
			\end{center}
			
			\vspace{1cm}
			
			\textbf{The proof}
			\begin{prooftree}
			«generateProof(output.leftParentOfBox)»
			«generateProof(output.rightParentOfBox)»
			\RightLabel{$ \sigma _«sigmaIndexMap.get(output.mgu)» $}
			\BinaryInfC{$\Box$}
			\end{prooftree}
			\vspace{1cm}
			\textbf{The substitutions}\\
			«FOR i:0..mguIndex-1»
				$ \sigma _{«i» = «LatexFormatter::format(indexSigmaMap.get(i))»}$\\
			«ENDFOR»
			
			\end{document}
		''' 
	}
	
	def private static String generateProof(Clause clause){
		System::out.print("Generating proof for: " + clause.toString);
		if(clause.leftParent == null && clause.rightParent == null){
			// axiom
			System::out.println(" which is a leaf.");
			System::out.print("FORMAT AXIOM: ");
			val String output = LatexFormatter::format(clause);
			System::out.println("\t\tLaTeX rep: " + output);
			return '''
			\AxiomC{}
			\RightLabel{ax}
			\UnaryInfC{ «output» }
			'''
		}else if(clause.leftParent != null && clause.rightParent == null){
			// factoring
			return '''
				«generateProof(clause.leftParent)»
				\RightLabel{ fac }
				\UnaryInfC{«LatexFormatter::format(clause)»}
			'''
		}else{
			// resolution
			System::out.println(" with parents: " + clause.leftParent.toString + ", " + clause.rightParent.toString);
			return '''
			«generateProof(clause.leftParent)»
			«generateProof(clause.rightParent)»
			\RightLabel{ res $ \sigma _{«sigmaIndexMap.get(clause.parentMGU)»} $}
			\BinaryInfC{«LatexFormatter::format(clause)»}
			'''
		}
	}
}