package app;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.util.EcoreUtil;

import util.FormulaFormatter;
import app.model.ClausalForm;
import app.model.Clause;
import at.cl.atp.resolution.AndFormula;
import at.cl.atp.resolution.Atom;
import at.cl.atp.resolution.Formula;
import at.cl.atp.resolution.ImplicationFormula;
import at.cl.atp.resolution.NegatedFormula;
import at.cl.atp.resolution.OrFormula;
import at.cl.atp.resolution.ResolutionFactory;
import at.cl.atp.resolution.ResolutionProblem;

public class FormulaTransformations {

	public static ClausalForm buildClausalForm(final Formula formula) {
		Formula f = normalize(formula);
		Formula cnf = toCNF(f);
		ClausalForm clausalForm = new ClausalForm();
		buildClausalForm(cnf, clausalForm, null);
		return clausalForm;
	}

	private static void buildClausalForm(final Formula formula, final ClausalForm clausalForm,
			final Clause currentClause) {
		if (formula instanceof Atom) {
			if (currentClause == null) {
				// single atom clause
				clausalForm.getClauses().add(Clause.createAtomClause((Atom) formula));
			} else {
				currentClause.getAtoms().add(((Atom) formula).getName());
			}
		} else if (formula instanceof NegatedFormula) {
			// we can only have negated ATOMS in a CNF
			if (currentClause == null) {
				// single negated atom clause
				clausalForm.getClauses().add(
						Clause.createNegatedAtomClause((Atom) ((NegatedFormula) formula).getInnerFormula()));
			} else {
				currentClause.getNegatedAtoms().add(((Atom) ((NegatedFormula) formula).getInnerFormula()).getName());
			}
		} else if (formula instanceof AndFormula) {
			buildClausalForm(((AndFormula) formula).getFirst(), clausalForm, null);
			buildClausalForm(((AndFormula) formula).getSecond(), clausalForm, null);
		} else if (formula instanceof OrFormula) {
			if (currentClause == null) {
				// create a new clause and add the children of this disjunction
				Clause clause = new Clause();
				clausalForm.getClauses().add(clause);
				buildClausalForm(((OrFormula) formula).getFirst(), clausalForm, clause);
				buildClausalForm(((OrFormula) formula).getSecond(), clausalForm, clause);
			} else {
				// remain in this clause and add children
				buildClausalForm(((OrFormula) formula).getFirst(), clausalForm, currentClause);
				buildClausalForm(((OrFormula) formula).getSecond(), clausalForm, currentClause);
			}
		}
	}

	public static Formula normalize(final Formula formula) {
		Formula f = implicationElimination(formula);
		f = doubleNegationElimination(f);
		System.out.println(FormulaFormatter.formulaToString(f));
		return pushNegationInside(f);
	}

	public static void normalize(final ResolutionProblem problem) {
		List<Formula> formulas = new ArrayList<Formula>(problem.getFormulas());
		for (Formula formula : formulas) {
			normalize(formula);
		}
	}

	public static Formula implicationElimination(final Formula formula) {
		System.out.println("->e: Search: " + FormulaFormatter.formulaToString(formula));
		if (formula instanceof Atom) {
			return formula;
		} else if (formula instanceof NegatedFormula) {
			implicationElimination(((NegatedFormula) formula).getInnerFormula());
			return formula;
		} else if (formula instanceof AndFormula) {
			implicationElimination(((AndFormula) formula).getFirst());
			implicationElimination(((AndFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof OrFormula) {
			implicationElimination(((OrFormula) formula).getFirst());
			implicationElimination(((OrFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof ImplicationFormula) {
			System.out.println("->e: Match: " + FormulaFormatter.formulaToString(formula));
			// replace the implication in the AST
			// with a disjunction where the first formula is negated
			ImplicationFormula implication = (ImplicationFormula) formula;
			implicationElimination(implication.getFirst());
			implicationElimination(implication.getSecond());
			OrFormula disjunction = ResolutionFactory.eINSTANCE.createOrFormula();
			NegatedFormula notLeft = ResolutionFactory.eINSTANCE.createNegatedFormula();
			notLeft.setInnerFormula(implication.getFirst());
			disjunction.setFirst(notLeft);
			disjunction.setSecond(implication.getSecond());
			if (implication.eContainer() instanceof Formula) {
				Formula parent = (Formula) implication.eContainer();
				replaceChild(parent, formula, disjunction);
				return disjunction;
			} else {
				ResolutionProblem parent = (ResolutionProblem) implication.eContainer();
				parent.getFormulas().remove(implication);
				parent.getFormulas().add(disjunction);
				return disjunction;
			}
		}
		return null;
	}

	public static Formula doubleNegationElimination(final Formula formula) {
		System.out.println("!!e: Search: " + FormulaFormatter.formulaToString(formula));
		if (formula instanceof Atom) {
			return formula;
		} else if (formula instanceof AndFormula) {
			doubleNegationElimination(((AndFormula) formula).getFirst());
			doubleNegationElimination(((AndFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof OrFormula) {
			doubleNegationElimination(((OrFormula) formula).getFirst());
			doubleNegationElimination(((OrFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof ImplicationFormula) {
			doubleNegationElimination(((ImplicationFormula) formula).getFirst());
			doubleNegationElimination(((ImplicationFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof NegatedFormula) {
			// check if the inner formula is an immediate negation
			NegatedFormula outerNegation = (NegatedFormula) formula;
			if (outerNegation.getInnerFormula() instanceof NegatedFormula) {
				if (outerNegation.eContainer() instanceof Formula) {
					System.out.println("!!e: Match: " + FormulaFormatter.formulaToString(outerNegation));
					NegatedFormula innerNegation = (NegatedFormula) outerNegation.getInnerFormula();
					Formula innerFormula = innerNegation.getInnerFormula();
					replaceChild((Formula) outerNegation.eContainer(), outerNegation, innerFormula);
					return innerFormula;
				} else if (outerNegation.eContainer() instanceof ResolutionProblem) {
					System.out.println("!!e: Match: " + FormulaFormatter.formulaToString(outerNegation));
					NegatedFormula innerNegation = (NegatedFormula) outerNegation.getInnerFormula();
					ResolutionProblem resolutionProblem = (ResolutionProblem) outerNegation.eContainer();
					resolutionProblem.getFormulas().remove(outerNegation);
					resolutionProblem.getFormulas().add(innerNegation.getInnerFormula());
					return innerNegation.getInnerFormula();
				}
			} else {
				doubleNegationElimination(outerNegation.getInnerFormula());
				return formula;
			}
		}
		return null;
	}

	public static Formula toCNF(final Formula formula) {
		System.out.println("CNF: Search: " + FormulaFormatter.formulaToString(formula));
		if (formula == null) {
			throw new RuntimeException("TOCNF: FORMULA IS NULL");
		}
		if (formula instanceof Atom) {
			return formula;
		} else if (formula instanceof AndFormula) {
			toCNF(((AndFormula) formula).getFirst());
			toCNF(((AndFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof OrFormula) {
			// if either child of the or is an and, we do a transformation
			OrFormula or = (OrFormula) formula;
			if (or.getFirst() instanceof AndFormula) {
				System.out.println("CNF: Match First: " + FormulaFormatter.formulaToString(or));
				// (P or (Q and R)) <-> ((P or Q) and (P or R))
				Formula p = or.getSecond();
				AndFormula and = (AndFormula) or.getFirst();
				Formula q = and.getFirst();
				Formula r = and.getSecond();
				OrFormula PorQ = ResolutionFactory.eINSTANCE.createOrFormula();
				OrFormula PorR = ResolutionFactory.eINSTANCE.createOrFormula();
				PorQ.setFirst(p);
				PorQ.setSecond(q);
				PorR.setFirst(EcoreUtil.copy(p));
				PorR.setSecond(r);
				AndFormula newAnd = ResolutionFactory.eINSTANCE.createAndFormula();
				newAnd.setFirst(PorQ);
				newAnd.setSecond(PorR);
				if (formula.eContainer() instanceof Formula) {
					replaceChild((Formula) formula.eContainer(), formula, newAnd);
					return toCNF((Formula) newAnd.eContainer());
				} else if (formula.eContainer() instanceof ResolutionProblem) {
					ResolutionProblem problem = (ResolutionProblem) formula.eContainer();
					problem.getFormulas().remove(formula);
					problem.getFormulas().add(newAnd);
					return toCNF(newAnd);
				}
			} else if (or.getSecond() instanceof AndFormula) {
				System.out.println("CNF: Match Second: " + FormulaFormatter.formulaToString(or));
				// (P or (Q and R)) <-> ((P or Q) and (P or R))
				Formula p = or.getFirst();
				AndFormula and = (AndFormula) or.getSecond();
				Formula q = and.getFirst();
				Formula r = and.getSecond();
				OrFormula PorQ = ResolutionFactory.eINSTANCE.createOrFormula();
				OrFormula PorR = ResolutionFactory.eINSTANCE.createOrFormula();
				PorQ.setFirst(p);
				PorQ.setSecond(q);
				PorR.setFirst(EcoreUtil.copy(p));
				PorR.setSecond(r);
				AndFormula newAnd = ResolutionFactory.eINSTANCE.createAndFormula();
				newAnd.setFirst(PorQ);
				newAnd.setSecond(PorR);
				if (formula.eContainer() instanceof Formula) {
					replaceChild((Formula) formula.eContainer(), formula, newAnd);
					return toCNF((Formula) newAnd.eContainer());
				} else if (formula.eContainer() instanceof ResolutionProblem) {
					ResolutionProblem problem = (ResolutionProblem) formula.eContainer();
					problem.getFormulas().remove(formula);
					problem.getFormulas().add(newAnd);
					return toCNF(newAnd);
				}
			} else {
				Formula f = toCNF(or.getFirst());
				if (f.eContainer() == or) {
					f = toCNF(or.getSecond());
					if (f.eContainer() instanceof Formula) {
						return (Formula) f.eContainer();
					} else {
						return f;
					}
				} else {
					if (f.eContainer() instanceof Formula) {
						return toCNF((Formula) f.eContainer());
					} else {
						return f;
					}
				}
			}
		} else if (formula instanceof ImplicationFormula) {
			return toCNF(implicationElimination(formula));
		} else if (formula instanceof NegatedFormula) {
			if (((NegatedFormula) formula).getInnerFormula() instanceof Atom == false) {
				return toCNF(pushNegationInside(formula));
			} else {
				return formula;
			}
		}
		// we can never reach this place anyways...
		throw new RuntimeException("Unknown formula: " + formula.getClass().getName());
	}

	public static Formula pushNegationInside(final Formula formula) {
		System.out.println("->!: Search: " + FormulaFormatter.formulaToString(formula));
		if (formula == null) {
			throw new RuntimeException("Formula is null");
		}
		if (formula instanceof Atom) {
			return formula;
		} else if (formula instanceof AndFormula) {
			pushNegationInside(((AndFormula) formula).getFirst());
			pushNegationInside(((AndFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof OrFormula) {
			pushNegationInside(((OrFormula) formula).getFirst());
			pushNegationInside(((OrFormula) formula).getSecond());
			return formula;
		} else if (formula instanceof ImplicationFormula) {
			Formula f = implicationElimination(formula);
			return pushNegationInside(f);
		} else if (formula instanceof NegatedFormula) {
			NegatedFormula outerNegation = (NegatedFormula) formula;
			Formula innerFormula = outerNegation.getInnerFormula();
			if (innerFormula instanceof Atom) {
				// we cannot push the negation further in
				return formula;
			} else if (innerFormula instanceof AndFormula) {
				System.out.println("->!: Match: " + FormulaFormatter.formulaToString(formula));
				// apply DeMorgan
				// !(a and b) == !a or !b
				AndFormula innerAnd = (AndFormula) innerFormula;
				Formula first = innerAnd.getFirst();
				Formula second = innerAnd.getSecond();
				OrFormula newOr = ResolutionFactory.eINSTANCE.createOrFormula();
				NegatedFormula negLeft = ResolutionFactory.eINSTANCE.createNegatedFormula();
				NegatedFormula negRight = ResolutionFactory.eINSTANCE.createNegatedFormula();
				negLeft.setInnerFormula(first);
				negRight.setInnerFormula(second);
				newOr.setFirst(negLeft);
				newOr.setSecond(negRight);
				if (outerNegation.eContainer() instanceof Formula) {
					Formula parent = (Formula) outerNegation.eContainer();
					replaceChild(parent, outerNegation, newOr);
					return pushNegationInside(parent);
				} else if (outerNegation.eContainer() instanceof ResolutionProblem) {
					ResolutionProblem problem = (ResolutionProblem) outerNegation.eContainer();
					problem.getFormulas().remove(formula);
					problem.getFormulas().add(newOr);
					return pushNegationInside(newOr);
				}
			} else if (innerFormula instanceof OrFormula) {
				System.out.println("->!: Match: " + FormulaFormatter.formulaToString(formula));
				// apply DeMorgan
				// !(a or b) == !a and !b
				OrFormula innerOr = (OrFormula) innerFormula;
				Formula first = innerOr.getFirst();
				Formula second = innerOr.getSecond();
				AndFormula newAnd = ResolutionFactory.eINSTANCE.createAndFormula();
				NegatedFormula negLeft = ResolutionFactory.eINSTANCE.createNegatedFormula();
				NegatedFormula negRight = ResolutionFactory.eINSTANCE.createNegatedFormula();
				negLeft.setInnerFormula(first);
				negRight.setInnerFormula(second);
				newAnd.setFirst(negLeft);
				newAnd.setSecond(negRight);
				if (outerNegation.eContainer() instanceof Formula) {
					Formula parent = (Formula) outerNegation.eContainer();
					replaceChild(parent, outerNegation, newAnd);
					return pushNegationInside(parent);
				} else if (outerNegation.eContainer() instanceof ResolutionProblem) {
					ResolutionProblem problem = (ResolutionProblem) outerNegation.eContainer();
					problem.getFormulas().remove(formula);
					problem.getFormulas().add(newAnd);
					return pushNegationInside(newAnd);
				}
			} else if (innerFormula instanceof ImplicationFormula) {
				Formula f = implicationElimination(formula);
				return pushNegationInside(f);
			} else if (innerFormula instanceof NegatedFormula) {
				// double negation elimination
				Formula f = doubleNegationElimination(formula);
				return pushNegationInside(f);
			}
		}
		throw new RuntimeException("Unreachable place reached!");
	}

	public static Formula merge(final ResolutionProblem problem) {
		AndFormula root = null;
		int problemSize = problem.getFormulas().size();
		if (problemSize == 1) {
			// nothing to merge with just 1 formula...
			return problem.getFormulas().get(0);
		}
		List<Formula> formulas = new ArrayList<Formula>(problem.getFormulas());
		for (int index = 0; index < problemSize; index++) {
			Formula currentFormula = formulas.get(index);
			if (root == null) {
				// create an AND
				root = ResolutionFactory.eINSTANCE.createAndFormula();
				root.setFirst(currentFormula);
			} else if (root.getSecond() == null) {
				// attach to lefthand-side of the AND
				root.setSecond(currentFormula);
			} else {
				// build another AND on top of root
				AndFormula newRoot = ResolutionFactory.eINSTANCE.createAndFormula();
				newRoot.setFirst(root);
				newRoot.setSecond(currentFormula);
				root = newRoot;
			}
		}
		return root;
	}

	public static void replaceChild(final Formula parent, final Formula childToReplace, final Formula newChild) {
		if (parent instanceof AndFormula) {
			if (((AndFormula) parent).getFirst() == childToReplace) {
				((AndFormula) parent).setFirst(newChild);
			} else if (((AndFormula) parent).getSecond() == childToReplace) {
				((AndFormula) parent).setSecond(newChild);
			}
		} else if (parent instanceof OrFormula) {
			if (((OrFormula) parent).getFirst() == childToReplace) {
				((OrFormula) parent).setFirst(newChild);
			} else if (((OrFormula) parent).getSecond() == childToReplace) {
				((OrFormula) parent).setSecond(newChild);
			}
		} else if (parent instanceof ImplicationFormula) {
			if (((ImplicationFormula) parent).getFirst() == childToReplace) {
				((ImplicationFormula) parent).setFirst(newChild);
			} else if (((ImplicationFormula) parent).getSecond() == childToReplace) {
				((ImplicationFormula) parent).setSecond(newChild);
			}
		} else if (parent instanceof NegatedFormula) {
			if (((NegatedFormula) parent).getInnerFormula() == childToReplace) {
				((NegatedFormula) parent).setInnerFormula(newChild);
			}
		}
		System.out.print("Validating...");
		FormulaValidator.validateFormulaIntegrity(rootFormula(parent));
		System.out.println(" check");
	}

	public static Formula rootFormula(final Formula formula) {
		Formula f = formula;
		while (f.eContainer() instanceof ResolutionProblem == false) {
			f = (Formula) f.eContainer();
		}
		return f;
	}
}
