package launcher;

import java.io.File;

import util.FormulaFormatter;
import util.ResolutionParser;
import app.FormulaTransformations;
import app.FormulaValidator;
import app.ResolutionProver;
import app.ResolutionResult;
import app.model.ClausalForm;
import at.cl.atp.resolution.Formula;
import at.cl.atp.resolution.NegatedFormula;
import at.cl.atp.resolution.ResolutionFactory;
import at.cl.atp.resolution.ResolutionProblem;

public class ResolutionProverLauncher {

	public static void main(final String[] args) {
		if (args.length != 1) {
			System.out.println("Please provide a *.res file containing the resolution problem.");
			System.exit(-1);
		}
		if (args[0].endsWith(".res") == false) {
			System.out.println("The provided file is not a *.res file.");
			System.exit(-1);
		}
		File problemFile = new File(args[0]);
		if (!problemFile.exists()) {
			System.out.println("Unable to locate file: " + problemFile.getAbsolutePath());
			System.exit(-1);
		}
		long systemStartTime = System.currentTimeMillis();
		ResolutionProblem problem = ResolutionParser.Parse(problemFile);
		long timeForLoad = System.currentTimeMillis() - systemStartTime;
		System.out.println(FormulaFormatter.formulaToString(problem));
		FormulaTransformations.normalize(problem);
		System.out.println(FormulaFormatter.formulaToString(problem));
		Formula merged = FormulaTransformations.merge(problem);
		System.out.println(FormulaFormatter.formulaToString(merged));
		FormulaValidator.validateFormulaIntegrity(merged);
		System.out.println("FORMULA INTEGRITY: VALID");
		NegatedFormula f = ResolutionFactory.eINSTANCE.createNegatedFormula();
		f.setInnerFormula(merged);
		problem.getFormulas().clear();
		problem.getFormulas().add(f);
		System.out.println("NEGATED FORMULA: " + FormulaFormatter.formulaToString(f));
		FormulaValidator.validateFormulaIntegrity(f);
		System.out.println("FORMULA INTEGRITY: VALID");
		long timeForCNF = System.currentTimeMillis() - timeForLoad - systemStartTime;
		ClausalForm clausalForm = FormulaTransformations.buildClausalForm(f);
		System.out.println(clausalForm.toString());
		ResolutionResult result = ResolutionProver.run(clausalForm);
		long timeForProof = System.currentTimeMillis() - timeForCNF - timeForLoad - systemStartTime;
		System.out.println("--------------------------------------------------------------------");
		System.out.println();
		System.out.println();
		System.out.println("___TIME__________________________________________");
		System.out.println("Time Load:  " + timeForLoad + "ms   (incl. EMF startup)");
		System.out.println("Time CNF:   " + timeForCNF + "ms");
		System.out.println("Time Proof: " + timeForProof + "ms");
		System.out.println("_________________________________________________");
		System.out.println("Time total: " + (timeForLoad + timeForCNF + timeForProof) + "ms");
		System.out.println();
		System.out.println();

		long memTotal = Runtime.getRuntime().totalMemory();
		long memMax = Runtime.getRuntime().maxMemory();
		long memUsed = memTotal - Runtime.getRuntime().freeMemory();
		System.out.println("___MEMORY________________________________________");
		System.out.println("Available: " + toMegabytes(memMax));
		System.out.println("Allocated: " + toMegabytes(memTotal));
		System.out.println("Used:      " + toMegabytes(memUsed));
		System.out.println();
		System.out.println();
		System.out.println("__RESULT_________________________________________");
		switch (result) {
		case BOX:
			System.out.println("The formula is VALID (resolution ended with BOX).");
			break;
		case EMPTY_SET:
			System.out.println("The formula is UNSATISFIABLE (resolution ended with EMPTY SET).");
			break;
		case SATISFIABLE:
			System.out.println("The formula is NOT VALID (resolution ended with remaining clauses).");
			break;
		}
	}

	private static String toMegabytes(final long bytes) {
		double megabytes = bytes;
		megabytes /= 1024;
		megabytes /= 1024;
		return String.format("%.2f MB", megabytes);
	}
}
